/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
#include "ATables.h" 

template <class VALTYPE>
void amodel<VALTYPE>::printTable(const char *filename) const
  // print amodel to file with the name filename (it'll be created or overwritten 
  // format : for a table : 
  //              aj j l m  val 
  // where aj is source word pos, j target word pos, l source sentence length, 
  //       m is target sentence length.
  //          
{
  if (is_distortion)
    cerr << "Dumping distortion table (d) to file:" << filename <<'\n';
  else 
    cerr << "Dumping alignment table (a) to file:" << filename <<'\n';
    

  ofstream of(filename);
  //  for(WordIndex l=0; l < MaxSentLength; l++) - Yaser
  //    for(WordIndex m=0;m<MaxSentLength;m++)
  for(WordIndex l=1; l < MaxSentLength; l++)
    for(WordIndex m=1;m<MaxSentLength;m++)
      if( is_distortion==0 )
	for(WordIndex j=1;j<=m; j++)
	  for(WordIndex i=0;i<=l; i++)
	    {
	      VALTYPE x=getValue(i, j, l, m);
	      if( x>PROB_SMOOTH )
		of << i << ' ' << j << ' ' << l << ' ' << m << ' ' << x << '\n';
	    }
      else
	for(WordIndex i=0;i<=l;i++)
	  for(WordIndex j=1;j<=m;j++)
	    
	    {
	      VALTYPE x=getValue(j, i, l, m);
	      if( x>PROB_SMOOTH )
		of << j << ' ' << i << ' ' << l << ' ' << m << ' ' << x << '\n';
	    }
}

template <class VALTYPE>
void amodel<VALTYPE>::normalize(amodel& aTable)const
{
  WordIndex i, j, l, m ;
  VALTYPE total;
  for(l=0;l<MaxSentLength;l++)
    for(m=0;m<MaxSentLength;m++)
      if( is_distortion==0 )
	for(j=1;j<=m; j++)
	  {
	    total=0.0;
	    for(i=0;i<=l;i++)
	      total+=getValue(i, j, l, m);
	    if( total )
	      for(i=0;i<=l;i++)
		aTable.getRef(i, j, l, m)=getValue(i, j, l, m)/total;
	  }
      else
	for(i=0;i<=l;i++)
	  {
	    total=0.0;
	    for(j=1;j<=m;j++)
	      total+=getValue(j, i, l, m);
	    if( total )
	      for(j=1;j<=m;j++)
		aTable.getRef(j, i, l, m)=getValue(j, i, l, m)/total;
	  }
}

template <class VALTYPE>
void amodel<VALTYPE>::readTable(const char *filename)
{
  /* This function reads the a table from a file.
     Each line is of the format:  aj j l m  val
     where aj is the source word position, j the target word position, 
     l the source sentence length, and m the target sentence length
     
     This function also works for a d table, where the positions
     of aj and i are swapped.  Both the a and d tables are 4 dimensional
     hashes; this function will simply read in the four values and keep
     them in that order when hashing the fifth value.
     NAS, 7/11/99
  */
  ifstream inf(filename);
  cerr << "Reading a/d table from " << filename << "\n";
  if(!inf){
    cerr << "\nCannot open " << filename<<"\n";
    exit(1);
  }
  WordIndex w, x, l, m;
  VALTYPE prob;
  while(inf >> w >> x >> l >> m >> prob )
    // the NULL word is added to the length
    // of the sentence in the tables, but discount it when you write the tables. 
    setValue(w, x, l, m, prob);
}

template class amodel<float> ; 


