/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
/* --------------------------------------------------------------------------*
 *                                                                           *
 * Module :ATables                                                           *
 *                                                                           *
 * Prototypes File: ATables.h                                                *
 *                                                                           *
 * Objective: Defines clases and methods for handling I/O for distortion  &  *
 *            alignment tables.                                              *
 *****************************************************************************/

#ifndef _atables_h
#define _atables_h 1

#include "defs.h"
//#include "logprob.h"
 
#include <assert.h>
 
#include <iostream>
#include <algorithm>
#include <functional>
#include <map>
#include <set>
//#include <vector>
#include "Vector.h"
#include <utility>
#include <hash_map>

#include <fstream.h>
#include "Array4.h"



/* ------------------- Class Defintions of amodel ---------------------------*/
/* Class Name: amodel:
   Objective: This defines the underlying data structure for distortiont prob.
   and count tables. They are defined as a hash table. Each entry in the hash
   table is the probability (d(j/l,m,i), where j is word target position, i is
   source word position connected to it, m is target sentence length, and l is
   source sentence length) or count collected for it. The  probability and the
   count are represented as log integer probability as 
   defined by the class LogProb .  

  This class is used to represents a Tables (probabiliity) and d (distortion)
  tables and also their corresponding count tables .
  
  *--------------------------------------------------------------------------*/

inline int Mabs(int a)
{
  if(a<0)
    return -a;
  else
    return a;
}

template <class VALTYPE>
class amodel
{
 private:
  Array4<VALTYPE> a;
  bool is_distortion ; 
  WordIndex MaxSentLength;
  bool ignoreL, ignoreM;
  short onlyDist;
  VALTYPE get(WordIndex aj, WordIndex j, WordIndex l, WordIndex m)const
    {
      assert( (!is_distortion) || aj<=m );assert( (!is_distortion) || j<=l );assert( (!is_distortion) || aj!=0 );
      assert( is_distortion    || aj<=l );assert( is_distortion    || j<=m );assert( (is_distortion) || j!=0 );
      assert( l<MaxSentLength );assert( m<MaxSentLength );
      if( onlyDist==3 )
	return a( (MaxSentLength+aj-j)/2 , 0, l, m);	
      else if( onlyDist==2 )
	return a( (MaxSentLength+aj-j)/2 , 0, MaxSentLength, (MaxSentLength+l-m)/2);	
      else if( onlyDist==1 )
	return a(Mabs(aj-j), 0, MaxSentLength, Mabs(l-m));
      else
	return a(aj, j, ignoreL?MaxSentLength:(l+1), ignoreM?MaxSentLength:(m+1));

    }
  
 public:
  amodel(bool flag, bool iL, bool iM, short oD, WordIndex maxlength=MAX_SENTENCE_LENGTH)
    : a(maxlength+1), is_distortion(flag), MaxSentLength(maxlength), ignoreL(iL), ignoreM(iM), onlyDist(oD)
    {}; 
  VALTYPE&getRef(WordIndex aj, WordIndex j, WordIndex l, WordIndex m)
    {
      assert( (!is_distortion) || aj<=m );assert( (!is_distortion) || j<=l );assert( (!is_distortion) || aj!=0 );
      assert( is_distortion    || aj<=l );assert( is_distortion    || j<=m );assert( (is_distortion) || j!=0 );
      assert( l<MaxSentLength );assert( m<MaxSentLength );
      if( onlyDist==3 )
	return a( (MaxSentLength+aj-j)/2 , 0, l, m);	
      else       if( onlyDist==2 )
	return a( (MaxSentLength+aj-j)/2 , 0, MaxSentLength, (MaxSentLength+l-m)/2);	
      else if( onlyDist==1 )
	return a(Mabs(aj-j), 0, MaxSentLength, Mabs(l-m));
      else
	return a(aj, j, ignoreL?MaxSentLength:(l+1), ignoreM?MaxSentLength:(m+1));
    }
  void setValue(WordIndex aj, WordIndex j, WordIndex l, WordIndex m, VALTYPE val)
    {
      getRef(aj, j, l, m)=val;
    }
  VALTYPE getValue(WordIndex aj, WordIndex j, WordIndex l, WordIndex m) const
    {
      return max(PROB_SMOOTH, get(aj, j, l, m));
    }
  void printTable(const char* filename)const ; 
  void normalize(amodel& aTable)const; 
  void readTable(const char *filename); 
  void clear()
    {
      static const VALTYPE zero=0.0;
      WordIndex i, j, l, m ;
      for(l=0; l < MaxSentLength; l++)
	for(m=0;m<MaxSentLength;m++)
	  if( is_distortion==0 )
	    for(j=1;j<=m; j++)
	      for(i=0; i <= l; i++)
		getRef(i, j, l, m)=zero;
	  else
	    for(i=0;i<=l;i++)
	      for(j=1;j<=m;j++)
		getRef(j, i, l, m)=zero;
    }
};

/* ------------------- End of amodel Class Definitions ----------------------*/

#endif
