/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
/*--
d4model: (currently) simplified representation of the IBM4 relative
         distortion probabilities
Franz Josef Och (7/August/99)
--*/
#ifndef _d4tables_h_define
#define _d4tables_h_define

class d4model
{
 private:
  Vector<pair<float, float> > d1, db1;
  WordIndex msl;
 public:
  d4model(int _msl)
    : d1(_msl*2+1), db1(_msl*2+1), msl(_msl)
    {}
  float&getCountRef_first(WordIndex j, WordIndex j_cp, int e_p, int fj)
    {
      assert(j>=1);
      assert(j_cp>=0);
      return d1[j-j_cp+msl].first;
    }
  float&getCountRef_bigger(WordIndex j, WordIndex j_prev, int fj)
    {
      assert(j>=1);
      assert(j_prev>=1);
      return db1[j-j_prev+msl].first;
    }
  float getProb_first(WordIndex j, WordIndex j_cp, int e_p, int fj)const
    {assert(j>=1);assert(j_cp>=0);
    assert(j<=msl);assert(j_cp<=msl);
    return max(PROB_SMOOTH, d1[j-j_cp+msl].second);}
  float getProb_second(WordIndex j, WordIndex j_prev, int e_p, int fj)const
    {assert(j>=1);assert(j_prev>=1);
    assert(j>j_prev);
    assert(j<=msl);assert(j_prev<=msl);
    return max(PROB_SMOOTH, db1[j-j_prev+msl].second);}
  void normalizeTable()
    {
      float sum=0.0;
      for(unsigned int i=0;i<d1.size();i++)sum+=d1[i].first;
      for(unsigned int i=0;i<d1.size();i++)d1[i].second=d1[i].first/sum;
      sum=0.0;
      for(unsigned int i=0;i<db1.size();i++)sum+=db1[i].first;
      for(unsigned int i=0;i<db1.size();i++)db1[i].second=db1[i].first/sum;
      cerr << "AFTER-NORMALIZE: " << *this << endl;
    }
  friend ostream&operator<<(ostream&out, const d4model&d4m)
    {
      out << "D1: \n";
      for(unsigned i=0;i<d4m.d1.size();i++)
	out << (int)(i)-(int)(d4m.msl) << ' ' << d4m.d1[i].second << '\n';
      out << endl;
      out << "D2: \n";
      for(unsigned i=0;i<d4m.db1.size();i++)
	out << (int)(i)-(int)(d4m.msl) << ' ' << d4m.db1[i].second << '\n';
      return out << endl;
    }
};

#endif
