/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
/*--
MoveSwapMatrix: Efficient representation for moving and swapping
around in IBM3 training.

Franz Josef Och (30/07/99)
--*/
#ifndef moveswap2_costs_h_defined
#define moveswap2_costs_h_defined
#include "alignment.h"
#include "transpair_model3.h"

template<class TRANSPAIR>
class MoveSwapMatrix : public alignment
{
 private:
  const TRANSPAIR&ef;
  const WordIndex l, m;
  Array2<LogProb, Vector<LogProb> > _cmove, _cswap;
  vector<int> changed;
  int changedCounter;
  const int modelnr;
 public:
  MoveSwapMatrix(const TRANSPAIR&_ef, const alignment&_a);
  void updateJ(WordIndex j, bool);
  void updateI(WordIndex i);
  void doMove(WordIndex _i, WordIndex _j);
  void doSwap(WordIndex _j1, WordIndex _j2);
  const TRANSPAIR&get_ef()const{return ef;}
  LogProb cmove(WordIndex x, WordIndex y)const
    {
#ifdef DEBUG
      assert(doubleEqual(_cmove(x, y), ef.scoreOfMove(*this, x, y)));
#endif
      return _cmove(x, y);
    }
  LogProb cswap(WordIndex x, WordIndex y)const
    {
      assert(y>x);
#ifdef DEBUG
      assert(doubleEqual(_cswap(x, y), ef.scoreOfSwap(*this, x, y)));
#endif
      return _cswap(x, y);
    }
  void printWrongs();
  bool isRight();
  //  template<class TRANSPAIR>
  friend ostream&operator<<(ostream&out, const MoveSwapMatrix<TRANSPAIR>&m)
    {return out << (alignment)m << "\nEF:\n"<< m.ef << "\nCMOVE\n"<<m._cmove << "\nCSWAP\n" << m._cswap << endl;};
};
#endif
