/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
#ifndef _ntables_h
#define _ntables_h 1
#include "Array2.h"
#include "Vector.h"
#include <assert.h>
//#include "logprob.h"
#include "defs.h"
#include "vocab.h"
template <class VALTYPE>
class nmodel
{
 private:
  Array2<VALTYPE, Vector<VALTYPE> > ntab;
 public:
  nmodel(int maxw, int maxn)
    : ntab(maxw, maxn, 0.0)
    {}
  VALTYPE getValue(int w, int n)const
    {
      if(n>=ntab.getLen2())
	return 0.0;
      else
	return max(float(ntab(w, n)), PROB_SMOOTH);
    }
  VALTYPE&getRef(int w, int n)
    {
      return ntab(w, n);
    }
  void normalize(nmodel&write)const
    {
      int h1=ntab.getLen1(), h2=ntab.getLen2();
      for(int i=0;i<h1;i++)
	{
	  VALTYPE sum=0.0;
	  for(int k=0;k<h2;k++)
	    sum+=getValue(i, k);
	  assert(sum);
	  for(int k=0;k<h2;k++)
	    write.getRef(i, k)=getValue(i, k)/sum;
	}
    }
  void clear()
    {
      int h1=ntab.getLen1(), h2=ntab.getLen2();
      for(int i=0;i<h1;i++)for(int k=0;k<h2;k++)
	ntab(i, k)=0;
    }
  //  void printNTable(int noEW, const char* filename) const;
  void printNTable(int noEW, const char* filename, 
		  const Vector<WordEntry>& evlist, bool) const;
  void readNTable(const char *filename);
  
};

#endif
