/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
/* Perplexity.h
 * ============
 * Mike Jahr, 7/15/99
 * Machine Translation group, WS99
 * Center for Language and Speech Processing
 * 
 * Last Modified by: Yaser Al-Onaizan, August 17, 1999
 *
 * Simple class used to calculate cross entropy and perplexity
 * of models.
 */

#ifndef _PERPLEXITY_H
#define _PERPLEXITY_H

#include <math.h>
#include <fstream.h>
#include "Vector.h"
#include "defs.h"
#include "Array2.h"
#include "Globals.h"

#define CROSS_ENTROPY_BASE 2

class Perplexity {
 private:
    double sum;
    int wc;
    static Array2<double, Vector<double> > *E_M_L;
    static const  int    initialized; // initialization flag
    static int Initialize();
    //    Vector<double> perp[4];
    //    Vector<double> ce[4];
    Vector<Vector<double> > perp;
    Vector<Vector<double> > ce;
    Vector<string> name ;
 public:
    Perplexity() {
      sum = 0 ;
      wc = 0;
      perp.clear();
      ce.clear();
      name.clear();
    }
    inline void clear() {
      sum = 0 ;
      wc = 0 ;
    }
    const size_t size()const{return(min(perp.size(), ce.size()));}
    inline void addFactor(const double p, const int count, const int l, 
			  const int m) {
      wc += count * m ; // number of french words 
      sum += count * (((*E_M_L)(l, m)) + p) ;
    }
    inline double perplexity() {
	return exp( -1*sum / wc);
    }

    inline double cross_entropy() {
      return (-1*sum / (log(CROSS_ENTROPY_BASE) * wc)); 
    }

    inline double word_count() {
	return wc;
    }
    
    inline double getSum(){
      return sum ;
    }

    void record(int model, unsigned int iteration);
    
    //    void writeLog(ofstream& of);
    friend void generatePerplexityReport(const Perplexity&, const Perplexity&, 
					 const Perplexity&, const Perplexity&, 
					 ostream&, int, int, bool); 
 private:
    string getModelName(int model){
      switch (model){
      case 0:
	return("1");
      case 1:
	return("2");
      case 2:
	return("2to3");
      case 3:
	return("3");
      case 4:
	return("3to4");
      case 5:
	return("4");
      default:
	return("unknown");
      }
    }
      
    void addModel(int model){
      perp.resize(model+1);
      ce.resize(model+1);
      name.resize(model+1);
      perp[model].clear();
      ce[model].clear();
      name[model] = getModelName(model);
    }  	
};


#endif
