/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
/*--
alignment: 'checked' alignment representation with automatic calculation
           of fertilities
Franz Josef Och (30/07/99)
--*/
#ifndef alignment_h_fjo_defined
#define alignment_h_fjo_defined
#include "Vector.h"
#include <assert.h>
#include "defs.h"

class alignment
{
 private:
  Vector<WordIndex> a;
  Vector<WordIndex> theHead;
  Vector<WordIndex> positionSum;
  Vector<WordIndex> f;
  WordIndex l;
  WordIndex m;
 public:
  alignment(int _l, int _m)
    : a(_m+1, (WordIndex)0), theHead(_l+1, (WordIndex)0), positionSum(_l+1, (WordIndex)0), f(_l+1, (WordIndex)0), l(_l), m(_m)
    {f[0]=m;}
  WordIndex get_l()const
    {return l;}
  WordIndex get_m()const
    {return m;}
  void set(WordIndex j, WordIndex aj)
    {
      unsigned int old_aj=a[j];
      assert(j<a.size());assert(aj<f.size());assert(old_aj<f.size());assert(f[old_aj]>0);
      positionSum[old_aj]-=j;
      if( old_aj&&theHead[old_aj]==j )
	{
	  theHead[old_aj]=0;
	  for(WordIndex nj=j+1;nj<=m;nj++)
	    if(a[nj]==old_aj)
	      {
		theHead[old_aj]=nj;
		break;
	      }
	}
      f[old_aj]--;
      positionSum[aj]+=j;
      if( theHead[aj]==0 || theHead[aj]>j )
	theHead[aj]=j;
      f[aj]++;
      a[j]=aj;
    }
  const Vector<WordIndex>& getAlignment() const 
    {return a ;}
  WordIndex get_al(WordIndex j)const
    {
      assert(j<a.size());
      return a[j];
    }
  WordIndex operator()(WordIndex j)const
    {
      assert(j<a.size());
      return a[j];
    }
  WordIndex fert(WordIndex i)const
    {
      assert(i<f.size());
      return f[i];
    }
  WordIndex get_head(WordIndex i)const
    {
      assert( theHead[i]==_get_head(i) );
      return theHead[i];
    }
  WordIndex get_center(WordIndex i)const
    {
      if( i==0 )return 0;
      assert((positionSum[i]+f[i])/f[i]==_get_center(i));
      return (positionSum[i]+f[i]-1)/f[i];
    }
  WordIndex _get_head(WordIndex i)const
    {
      if( fert(i)==0 )return 0;
      for(WordIndex j=1;j<=m;j++)
	if( a[j]==i )
	  return j;
      return 0;
    }
  WordIndex _get_center(WordIndex i)const
    {
      if( i==0 )return 0;
      assert(fert(i));
      WordIndex sum=0;
      for(WordIndex j=1;j<=m;j++)
	if( a[j]==i )
	  sum+=j;
      return (sum+fert(i)-1)/fert(i);
    }
  WordIndex prev_cept(WordIndex i)const
    {
      if( i==0 )return 0;
      WordIndex k=i-1;
      while(k&&fert(k)==0)
	k--;
      return k;
    }
  WordIndex next_cept(WordIndex i)const
    {
      WordIndex k=i+1;
      while(k<l+1&&fert(k)==0)
	k++;
      return k;
    }
  WordIndex prev_in_cept(WordIndex j)const
    {
      WordIndex k=j-1;
      while(k&&a[k]!=a[j])
	k--;
      assert(k);
      return k;
    }
  friend ostream &operator<<(ostream&out, const alignment&a);
  friend bool operator==(const alignment&a, const alignment&b)
    {
      assert(a.a.size()==b.a.size());
      for(WordIndex j=1;j<a.a.size();j++)
	if(a(j)!=b(j))
	  return 0;
      return 1;
    }
  friend bool operator<(const alignment&x, const alignment&y)
    {
      assert(x.get_m()==y.get_m());
      for(WordIndex j=1;j<x.get_m();j++)
	if( x(j)<y(j) ) 
	  return 1;
	else 
	  if( y(j)<x(j) )
	    return 0;
      return 0;
    }
  friend int differences(const alignment&x, const alignment&y){
    int count=0;
    assert(x.get_m()==y.get_m());
    for(WordIndex j=1;j<x.get_m();j++)
      count += (x(j)!=y(j));
    return count;
  }
};
  
#endif
