/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
/*--
collCounts: collects counts over the neighborhood of one
or a vector of alignments. 
IMPORTANT: double-counting of alignments is not avoided!

Franz Josef Och (30/07/99)
--*/
#include "alignment.h"
#include "transpair_model3.h"
#include <map>
#include "collCounts.h"
#include "MoveSwapMatrix.h"
// unifies collectCountsOverAlignments and findAlignmentNeighborhood FJO-20/07/99
template<class TRANSPAIR>
void collectCountsOverNeighborhood(const MoveSwapMatrix<TRANSPAIR>&msc, LogProb ascore, Array2<LogProb, Vector<LogProb> >&dtcount, Array2<LogProb, Vector<LogProb> >&ncount, LogProb&p1count, LogProb&p0count, LogProb&total_count)
{
  const WordIndex l=msc.get_l(), m=msc.get_m();
  Array2<LogProb, Vector<LogProb> > cmove(l+1, m+1), cswap(l+1, m+1);
  Vector<LogProb> negmove(m+1), negswap(m+1), plus1fert(l+1), minus1fert(l+1);
  LogProb total_move=ascore, total_swap;
  for(WordIndex j=1;j<=m;j++)for(WordIndex i=0;i<=l;i++)if( msc(j)!=i/*&&msc.get_ef().exists(i, j)*/ )
    {
      LogProb newscore=ascore*msc.cmove(i, j);
      total_move+=newscore;
      cmove(i, j)+=newscore;
      negmove[j]+=newscore;
      plus1fert[i]+=newscore;
      minus1fert[msc(j)]+=newscore;
    }
  for(WordIndex j1=1;j1<=m;j1++)for(WordIndex j2=j1+1;j2<=m;j2++)if( msc(j1)!=msc(j2)/*&&msc.get_ef().exists(msc(j1), j2)&&msc.get_ef().exists(msc(j2), j1)*/ )
    {
      LogProb newscore=ascore*msc.cswap(j1, j2);
      total_swap+=newscore;
      cswap(msc(j1), j2)+=newscore;
      cswap(msc(j2), j1)+=newscore;
      negswap[j1]+=newscore; 
      negswap[j2]+=newscore;
    }
  total_count+=total_move+total_swap;
  for(WordIndex j=1;j<=m;j++)
    for(WordIndex i=0;i<=l;i++)
      dtcount(i, j) += ((i==msc(j)) ? (total_count-(negmove[j]+negswap[j])) : (cswap(i, j)+cmove(i, j)));
  for(WordIndex i=0;i<=l;i++)
    {
      assert( msc.fert(i)<MAX_FERTILITY );
      LogProb temp=minus1fert[i]+plus1fert[i];
      ncount(i, msc.fert(i))+=total_count-temp;
      if(msc.fert(i)>0)
	ncount(i, msc.fert(i)-1)+=minus1fert[i];
      else
	if( minus1fert[i]!=0.0 )cerr << "ERROR: M1Fa: " << minus1fert[i] << endl;
      if(msc.fert(i)+1<MAX_FERTILITY) 
	ncount(i, msc.fert(i)+1)+=plus1fert[i];
    }
  LogProb temp=minus1fert[0]+plus1fert[0];
  p1count += (total_count-temp)*(LogProb)msc.fert(0);
  p0count += (total_count-temp)*(LogProb)(m-2*msc.fert(0));
  assert( m-2*msc.fert(0)>=0);
  if( msc.fert(0)>0 )
    {
      p1count += (minus1fert[0])*(LogProb)(msc.fert(0)-1);
      p0count += (minus1fert[0])*(LogProb)(m-2*(msc.fert(0)-1));
    }
  else
    if( minus1fert[0]!=0.0 )cerr << "ERROR: M1Fb: " << minus1fert[0] << endl;
  if(m-2*(msc.fert(0)+1)>=0)
    {
      p1count += (plus1fert[0])*(LogProb)(msc.fert(0)+1);
      p0count += (plus1fert[0])*(LogProb)(m-2*(msc.fert(0)+1));
    }
  else
    if( plus1fert[0]!=0.0 )cerr << "ERROR: P1Fb: " << plus1fert[0] << endl;
};

template<class TRANSPAIR>
void collectCountsOverNeighborhoodForD4(const MoveSwapMatrix<TRANSPAIR>&msc, LogProb ascore, d4model*d4Table)
{
  const WordIndex m=msc.get_m();
  for(WordIndex j=1;j<=m;++j)
    if( msc(j)!=0 )
      if( msc.get_head(msc(j))==j)
	d4Table->getCountRef_first(j,msc.get_center(msc.prev_cept(msc(j))),0,0)+=1.0;
      else
	d4Table->getCountRef_bigger(j,msc.prev_in_cept(j),0)+=1.0;
}

template<class TRANSPAIR>
void collectCountsOverNeighborhood(const Vector<pair<MoveSwapMatrix<TRANSPAIR>*, LogProb> >&smsc, Vector<WordIndex>&es, Vector<WordIndex>&fs, tmodel<COUNT, PROB>&tTable, amodel<COUNT>&aCountTable, amodel<COUNT>&dCountTable, nmodel<COUNT>&nCountTable, double&p1count, double&p0count, LogProb&_total, int count, bool addCounts, d4model*d4Table)
{
#ifdef DEBUG_TRICKY_IBM3
  cout << "EF:"<<ef;
#endif
  const WordIndex l=es.size()-1, m=fs.size()-1;
  Array2<LogProb, Vector<LogProb> > dtcount(l+1, m+1), ncount(l+1, MAX_FERTILITY+1);
  LogProb p0, p1, total=0;
  for(unsigned int i=0;i<smsc.size();++i)
    {
      collectCountsOverNeighborhood(*smsc[i].first, smsc[i].second, dtcount, ncount, p1, p0, total);
      if( addCounts && d4Table )
	collectCountsOverNeighborhoodForD4(*smsc[i].first, smsc[i].second, d4Table);
    }
  _total=total;
  total/=(double)count;
#ifdef DEBUG_TRICKY_IBM3
  cout << "DT:" << dtcount << endl;
  cout << "NT:" << ncount << endl;
  cout << "P0:"<<p0 << "  P1:"<<p1 << endl;
  cout << "TOTAL: " << total << endl;
#endif
  if( addCounts )
    {
      for(WordIndex i=0;i<=l;i++) 
	{
	  for(WordIndex j=1;j<=m;j++)
	    {
	      LogProb ijadd=dtcount(i, j)/total;
	      if( ijadd>COUNTINCREASE_CUTOFF )
		{
		  tTable.incCount(es[i], fs[j], ijadd);
		  dCountTable.getRef(j, i, l, m)+=ijadd;
		  aCountTable.getRef(i, j, l, m)+=ijadd;
		}
	    }
	  for(WordIndex n=0;n<MAX_FERTILITY;n++)
	    nCountTable.getRef(es[i], n)+=ncount(i, n)/total;
	}
      p0count+=p0/total;
      p1count+=p1/total;
    }
}


template void collectCountsOverNeighborhood(const Vector<pair<MoveSwapMatrix<transpair_model3>*, LogProb> >&smsc, Vector<WordIndex>&es, Vector<WordIndex>&fs, tmodel<COUNT, PROB>&tTable, amodel<COUNT>&aCountTable, amodel<COUNT>&dCountTable, nmodel<COUNT>&nCountTable, double&p1count, double&p0count, LogProb&_total, int count, bool addCounts, d4model*d4Table);
template void collectCountsOverNeighborhood(const Vector<pair<MoveSwapMatrix<transpair_model4>*, LogProb> >&smsc, Vector<WordIndex>&es, Vector<WordIndex>&fs, tmodel<COUNT, PROB>&tTable, amodel<COUNT>&aCountTable, amodel<COUNT>&dCountTable, nmodel<COUNT>&nCountTable, double&p1count, double&p0count, LogProb&_total, int count, bool addCounts, d4model*d4Table);
