/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
/* --------------------------------------------------------------------------*
 *                                                                           *
 * Module : getSentece                                                       *
 *                                                                           *
 * Method Definitions File: getSentence.cc                                   *
 *                                                                           *
 * Objective: Defines clases and methods for handling I/O for the parallel   *
 *            corpus.                                                        *
 *****************************************************************************/


#include "getSentence.h"
#include <iostream>
#include <strstream>

/* -------------- Method Defnitions for Class sentenceHandler ---------------*/

/*sentenceHandler::sentenceHandler(const char*  filename)
  // This method is the constructor of the class, it also intitializes the 
  // sentence pair sequential number (count) to zero.

{
  inputFilename = filename ;
  inputFile.open(filename);
  if(!inputFile){
    cerr << "\nCannot open " << filename;
    exit(1);
  }
  currentSentence = 0;
  totalPairs = 0 ;
  noSentInBuffer = 0 ;
  Buffer.clear();
  //  Buffer.reserve(TRAIN_BUFFER_SIZE);
}*/

sentenceHandler::sentenceHandler(const char*  filename, vcbList* elist=0, 
				 vcbList*  flist=0)
  // This method is the constructor of the class, it also intitializes the 
  // sentence pair sequential number (count) to zero.

{
  readflag = false ;
  allInMemory = false ;
  inputFilename = filename ;
  inputFile.open(filename);
  inputFile.close();
  pair_no = 0 ;
  inputFile.open(filename);
  if(!inputFile){
    cerr << "\nCannot open " << filename;
    exit(1);
  }
  currentSentence = 0;
  totalPairs = 0 ;
  pair_no = 0 ;
  noSentInBuffer = 0 ;
  Buffer.clear();
  if (elist && flist){
    cerr << "Calculating vocabulary frequencies from corpus " << filename << '\n';
    sentPair s ;
    while (getNextSentence(s, elist, flist)){totalPairs++;}
  //  Buffer.reserve(TRAIN_BUFFER_SIZE);
  }
}

void sentenceHandler::rewind()
{
  currentSentence = 0;
  readflag = false ;
  if (!allInMemory || 
      !(Buffer.size() >= 1 && Buffer[currentSentence].sentenceNo == 1)){
    // check if the buffer doe not already has the first chunk of pairs 
    cerr << "Actually rewinding " << Buffer.size() << ' ' << currentSentence;
    if (Buffer.size() > 0)
      cerr << ' ' <<  Buffer[currentSentence].sentenceNo << '\n';
    //    totalPairs = 0 ;
    pair_no = 0 ;
    noSentInBuffer = 0 ;
    Buffer.clear();
  }
  if (!allInMemory){
    inputFile.close(); // rewind file stream reader ;
    inputFile.open(inputFilename);
    if(!inputFile){
      cerr << "\nCannot open " << inputFilename;
      exit(1);
    }
  }
}

  
bool sentenceHandler::getNextSentence(sentPair& sent, vcbList* elist=0, vcbList* flist=0)
{
  sentPair s ;
  if (readflag){
    cerr << "Attempting to read from the end of corpus, rewinding\n";
    rewind();
    return(false);
  } 
  if (currentSentence >= noSentInBuffer){
    if (allInMemory)
      return(false);
    /* no more sentences in buffer */
    noSentInBuffer = 0 ;
    currentSentence = 0 ;
    Buffer.clear();
    cerr << "Reading more sentence pairs into memory ... \n";
    while((noSentInBuffer < TRAIN_BUFFER_SIZE) && readNextSentence(s)){
      if ((s.fSent.size()-1) > (MAX_FERTILITY-1) * (s.eSent.size()-1)){
	cerr << "WARNING: The following sentence pair has source/target sentence length ration more than\n"<<
	  "the maximum allowed limit for a source word fertility\n"<<
	  " source length = " << s.eSent.size()-1 << " target length = " << s.fSent.size()-1 <<
	  " ratio " << double(s.fSent.size()-1)/  (s.eSent.size()-1) << " ferility limit : " <<
	  MAX_FERTILITY-1 << '\n';
	cerr << "Ignoring sentence \n";
	s.printSentencePair(cerr);
      }
      else {	  
	Buffer.push_back(s) ;
	if (elist && flist){
	  if ((*elist).size() > 0){
	    for (WordIndex i= 0 ; i < s.eSent.size() ; i++){
	      if (s.eSent[i] >= (*elist).uniqTokens()){
		cerr << "ERROR: source word " << s.eSent[i] << " is not in the vocabulary list \n";
		exit(-1);
	      }
	      (*elist).incFreq(s.eSent[i], s.noOccurrences);
	    }
	    (*elist).incFreq(0, s.noOccurrences); // increment count for NULL word 
	  }
	  if ((*flist).size() > 0)
	    for (WordIndex j= 1 ; j < s.fSent.size() ; j++){
	      if (s.fSent[j] >= (*flist).uniqTokens()){
		cerr << "ERROR: target word " << s.fSent[j] << " is not in the vocabulary list \n";
		exit(-1);
	      }
	      (*flist).incFreq(s.fSent[j], s.noOccurrences);
	    }
	}
	noSentInBuffer++;
      }
    }
    if (inputFile.eof()){
      cerr << " Reached end of corpus file ... \n";
      allInMemory = (Buffer.size() >= 1 && 
		     Buffer[currentSentence].sentenceNo == 1) ;
      if (allInMemory)
	cerr << "Corpus fits in memory, corpus has: " << Buffer.size() <<
	  " sentence pairs.\n";
    }
  }
  if(noSentInBuffer <= 0 ){
    cerr << "# sent in buffer " << noSentInBuffer << '\n';
    readflag = true ;
    return(false);
  }
  sent = Buffer[currentSentence++] ;
  return true ;
}
bool sentenceHandler::readNextSentence(sentPair& sent)
  /* This method reads in a new pair of sentences, each pair is read from the 
     corpus file as line triples. The first line the no of times this line 
     pair occured in the corpus, the second line is the source sentence and 
     the third is the target sentence. The sentences are represented by a space
     separated positive integer token ids. */
{

  string line;
  bool fail(false) ;
  
  sent.clear();
  if (getline(inputFile, line)){
    istrstream buffer(line.c_str());
    buffer >> sent.noOccurrences ;
  }
  else {
    fail = true ;;
  }
  if (getline(inputFile, line)){
    istrstream buffer(line.c_str());
    WordIndex w;  // w is a local variabe for token id
    sent.eSent.push_back(0); // each source word is assumed to have 0 == 
    // a null word (id 0) at the begining of the sentence. 
    while(buffer>>w){ // read source sentece , word by word .
      if (sent.eSent.size() < MAX_SENTENCE_LENGTH)
	sent.eSent.push_back(w);
      else {
	cerr << "ERROR: getSentence.cc:getNextSentence(): sentence exceeds preset length limit of : " << MAX_SENTENCE_LENGTH << '\n';
	cerr << "The following sentence will be truncated\n" << line;
	break ;
      }
    }
  }
  else {
    fail = true ;
  }
  if (getline(inputFile, line)){
    istrstream buffer(line.c_str());
    WordIndex w;  // w is a local variabe for token id
    sent.fSent.push_back(0); //0 is inserted for program uniformity
    while(buffer>>w){ // read target sentece , word by word .
      if (sent.fSent.size() < MAX_SENTENCE_LENGTH)
	sent.fSent.push_back(w);
      else {
	cerr << "ERROR: getSentence.cc:getNextSentence(): sentence exceeds preset length limit of : " << MAX_SENTENCE_LENGTH << '\n';
	cerr << "The following sentence will be truncated\n" << line;
	break ;
      }
    }
  }
  else {
    fail = true ;
  }
  if (fail){
    sent.eSent.clear();
    sent.fSent.clear();
    sent.sentenceNo = 0 ;
    sent.noOccurrences = 0 ;
    return(false);
  }  
  sent.sentenceNo = ++pair_no;
  if(pair_no % 10000 == 0)  // This is only to print a verbose message
                                // in the stderr, for every 1000 sentece pairs
    cerr << endl << "Just read sentence pair no:"  << sent.sentenceNo  
	 << " , source length: " << sent.eSent.size()-1 
	 << " , target length: " << sent.fSent.size()-1 <<"\n";
  return true;
}


/* ------------- End of Method Definition of Class sentenceHandler ----------*/
