/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
#include <strstream>
#include "getSentence.h"
#include "TTables.h"
#include "model1.h"
#include "model2.h"
#include "model3.h"
#include "file_spec.h"
#include "defs.h"
#include "vocab.h"
#include "Perplexity.h"
#include "Dictionary.h"
#include "utility.h" 

// Global Variables 

float PROB_CUTOFF=1e-7, COUNTINCREASE_CUTOFF=1e-7, PROB_SMOOTH=1e-7;

int ADTableIgnoreL=0, ADTableIgnoreM=0, ADTableSpecial=0;

bool Verbose, Log, Peg, Transfer, Transfer2to3, useDict = false  ;

string Prefix, LogFilename, OPath, LogProbPath, Usage, 
  SourceVocabFilename, TargetVocabFilename, CorpusFilename, 
  TestCorpusFilename, t_Filename, a_Filename, p0_Filename, d_Filename, 
  n_Filename, dictionary_Filename;

ofstream logmsg ;

int Model1_Iterations, Model2_Iterations, Model3_Iterations, Model4_Iterations ;
int Model1_Dump_Freq, Model2_Dump_Freq, Model3_Dump_Freq, Transfer_Dump_Freq, Model4_Dump_Freq ;
unsigned int MAX_SENTENCE_LENGTH ;
double P0 ;

bool FIX_P0, NODUMPS  ;


void initGlobals(void)
{
  // initialize global variables 
  // set default values of options 

  P0 = 0 ;
  FIX_P0 = false ;
  NODUMPS = false ;
  Verbose = false ;
  Log = false ;
  Prefix = Get_File_Spec();
  LogFilename= Prefix + ".log";
  Transfer = TRANSFER ;
  LogProbPath = "./" ;
  Model1_Iterations = MAX_ITERATIONS_MODEL1 ;
  Model2_Iterations = MAX_ITERATIONS_MODEL2 ;
  Model3_Iterations = MAX_ITERATIONS_MODEL3 ;
  Model4_Iterations = MAX_ITERATIONS_MODEL4 ;
  Model1_Dump_Freq = TABLE_DUMP_FREQ_1 ;
  Model2_Dump_Freq = TABLE_DUMP_FREQ_2 ;
  Model3_Dump_Freq = TABLE_DUMP_FREQ_3 ;
  Model4_Dump_Freq = 1 ;
  Transfer_Dump_Freq = TABLE_DUMP_FREQ_T ;
  Transfer2to3 = true;
  MAX_SENTENCE_LENGTH = MAX_SENTENCE_LENGTH_ALLOWED ;
}


int main(int argc, char* argv[])
{

  time_t st1, fn;
  st1 = time(NULL);                    // starting time

  int lastmodel = 0 ;

  string temp(argv[0]);
  Usage = temp + " original_text_fname trans_file1 [trans_file2 ....]\n";
  if(argc < 2){
    printHelp();    
    exit(1);
  }

  initGlobals() ;
  parseArguments(argc, argv);

  if (Log)
    logmsg.open(LogFilename.c_str(), ios::out);

  printOptions(cerr);
  string repFilename = Prefix + ".rep" ;
  ofstream of2(repFilename.c_str());
  printOptions(of2) ;
  Perplexity trainPerp, testPerp, trainViterbiPerp, testViterbiPerp ;

  cerr << "reading vocabulary files \n";
  vcbList eTrainVcbList(SourceVocabFilename.c_str()), fTrainVcbList(TargetVocabFilename.c_str());
  eTrainVcbList.readVocabList();
  fTrainVcbList.readVocabList();
  cerr << "Source vocabulary list has " << eTrainVcbList.uniqTokens() << " unique tokens \n";
  cerr << "Target vocabulary list has " << fTrainVcbList.uniqTokens() << " unique tokens \n";
  
  of2 << "Source vocabulary list has " << eTrainVcbList.uniqTokens() << " unique tokens \n";
  of2 << "Target vocabulary list has " << fTrainVcbList.uniqTokens() << " unique tokens \n";

  vcbList eTestVcbList(eTrainVcbList) ;
  vcbList fTestVcbList(fTrainVcbList) ;
  
  sentenceHandler* testCorpus=0, corpus(CorpusFilename.c_str(), &eTrainVcbList, &fTrainVcbList);

  if (TestCorpusFilename == "NONE")
    TestCorpusFilename = "";

  if (TestCorpusFilename != ""){
    cerr << "Test corpus will be read from: " << TestCorpusFilename << '\n';
      testCorpus= new sentenceHandler(TestCorpusFilename.c_str(), 
						       &eTestVcbList, &fTestVcbList);
      cerr << " Test total # sentence pairs : " <<(*testCorpus).getTotalNoPairs() <<'\n';

      cerr << "Size of the source portion of test corpus: " << eTestVcbList.totalVocab() << " tokens\n";
      cerr << "Size of the target portion of test corpus: " << fTestVcbList.totalVocab() << " tokens \n";
      cerr << "In source portion of the test corpus, only " << eTestVcbList.uniqTokensInCorpus() << " unique tokens appeared\n";
      cerr << "In target portion of the test corpus, only " << fTestVcbList.uniqTokensInCorpus() << " unique tokens appeared\n";
      cerr << "ratio (target/source) : " << double(fTestVcbList.totalVocab()) /
	eTestVcbList.totalVocab() << '\n';
      of2 << "Size of the source portion of test corpus: " << eTestVcbList.totalVocab() << " tokens\n";
      of2 << "Size of the target portion of test corpus: " << fTestVcbList.totalVocab() << " tokens \n";
      of2 << "In source portion of the test corpus, only " << eTestVcbList.uniqTokensInCorpus() << " unique tokens appeared\n";
      of2 << "In target portion of the test corpus, only " << fTestVcbList.uniqTokensInCorpus() << " unique tokens appeared\n";
      of2 << "ratio (target/source) : " << double(fTestVcbList.totalVocab()) /
	eTestVcbList.totalVocab() << '\n';
  }
  
  cerr << " Train total # sentence pairs : " << corpus.getTotalNoPairs() << '\n';
  cerr << "Size of source portion of the training corpus: " << eTrainVcbList.totalVocab() << " tokens\n";
  cerr << "Size of the target portion of the training corpus: " << fTrainVcbList.totalVocab() << " tokens \n";
  cerr << "In source portion of the training corpus, only " << eTrainVcbList.uniqTokensInCorpus() << " unique tokens appeared\n";
  cerr << "In target portion of the training corpus, only " << fTrainVcbList.uniqTokensInCorpus() << " unique tokens appeared\n";

  cerr << "ratio (target/source) : " << double(fTrainVcbList.totalVocab()) /
    eTrainVcbList.totalVocab() << '\n';
  of2 << "Size of source portion of the training corpus: " << eTrainVcbList.totalVocab() << " tokens\n";
  of2 << "Size of the target portion of the training corpus: " << fTrainVcbList.totalVocab() << " tokens \n";
  of2 << "In source portion of the training corpus, only " << eTrainVcbList.uniqTokensInCorpus() << " unique tokens appeared\n";
  of2 << "In target portion of the training corpus, only " << fTrainVcbList.uniqTokensInCorpus() << " unique tokens appeared\n";
  
  of2 << "ratio (target/source) : " << double(fTrainVcbList.totalVocab()) /
    eTrainVcbList.totalVocab() << '\n';

  // load dictionary
  Dictionary *dictionary;  
  if (useDict) dictionary = new Dictionary(dictionary_Filename.c_str());
  else dictionary = new Dictionary("");
  
  // initialize model1
  model1 m1(CorpusFilename.c_str(), eTrainVcbList, fTrainVcbList);
  bool seedModel1 = false ;
   if(Model1_Iterations > 0){
     if (t_Filename != "NONE" && t_Filename != ""){
       seedModel1 = true ;
       m1.load_table(t_Filename.c_str());
     }
     // run the EM alg on it
     lastmodel = 1;
     if (testCorpus)
       m1.em_with_tricks(Model1_Iterations, trainPerp, corpus, seedModel1, 
			 *dictionary, useDict, &testPerp, testCorpus, 
			 trainViterbiPerp, &testViterbiPerp);
     else 
       m1.em_with_tricks(Model1_Iterations,trainPerp,corpus,seedModel1,*dictionary, useDict, 0,0,trainViterbiPerp,0);
   }
   else{
     if (Model2_Iterations > 0 || Model3_Iterations > 0 || Transfer2to3)
       // read in model1 output file (t_Filename)
       m1.load_table(t_Filename.c_str());
   }
   
   // transfer tTable from model1 to model2
   
   model2 m2(m1);
   
   if(Model2_Iterations > 0){
     // run EM alg on tTable and aTable
     lastmodel = 2;
     m2.initialize_table_uniformly();
     if (testCorpus)
       m2.em_with_tricks(m1.getTTable(), Model2_Iterations, trainPerp, corpus, 
			 &testPerp, testCorpus, trainViterbiPerp, testViterbiPerp);
     else 
       m2.em_with_tricks(m1.getTTable(), Model2_Iterations,trainPerp,corpus,0,0,trainViterbiPerp,
			 testViterbiPerp);
   }
   else{
     if (Transfer2to3 || Model3_Iterations > 0){
       // read in model2 output files (a_Filename)
       cerr << "Transfer m2 load table is called \n";
       m2.load_table(a_Filename.c_str());
     }
   }
   
   model3 m3(m1, m2);  // If we're not supposed to do this transfer, should this code still execute?
   
   if(Transfer2to3){

     // transfer model2 to model3
     lastmodel = 3 ;
     string test_alignfile = Prefix +".tst.A2to3";
     if (testCorpus)
       m2.em_loop(2, m1.getTTable(), 1, testPerp, *testCorpus, 
		  Transfer_Dump_Freq==1&&!NODUMPS, 
		  test_alignfile.c_str(), testViterbiPerp, true);
     if (Transfer == TRANSFER_SIMPLE)
       m3.transferSimple(m1, m2, corpus, Transfer_Dump_Freq==1&&!NODUMPS, 
			 trainPerp, trainViterbiPerp);
     else  
       m3.transfer(m1, m2, corpus, Transfer_Dump_Freq==1&&!NODUMPS, 
		   trainPerp, trainViterbiPerp);

     if (testCorpus)
       cerr << "\nTransfer: TEST CROSS-ENTROPY " << testPerp.cross_entropy()
	    << " PERPLEXITY " << testPerp.perplexity() << "\n\n";
   }
   else{
     if (Model3_Iterations > 0 || Model4_Iterations > 0){
       cerr << "model3 iterations " << Model3_Iterations<<'\n';
       // read in model2to3 output files (p0_Filename, n_filename, and d_filename)
       m3.load_tables(n_Filename.c_str(), d_Filename.c_str(), p0_Filename.c_str());
     }
   }

   if(Model3_Iterations > 0){
     lastmodel = 4; 
     if (testCorpus)
       m3.viterbi(m1.getTTable(), m2.getATable(), m2.getACountTable(), 
		  Model3_Iterations, Model4_Iterations, trainPerp, 
		  trainViterbiPerp, corpus, &testPerp, &testViterbiPerp, testCorpus);
     else 
       m3.viterbi(m1.getTTable(),m2.getATable(),m2.getACountTable(),
		  Model3_Iterations,Model4_Iterations,trainPerp,trainViterbiPerp,corpus,0,0,0);
   }

   if (!NODUMPS){
     string t_inv_file = Prefix + ".ti.final" ;
     m1.getTTable().printProbTableInverse(t_inv_file.c_str(), m1.getEnglishVocabList(), 
					  m1.getFrenchVocabList(), 
					  m1.getETotalWCount(), 
					  m1.getFTotalWCount());
     t_inv_file = Prefix + ".actual.ti.final" ;
     m1.getTTable().printProbTableInverse(t_inv_file.c_str(), 
					  eTrainVcbList.getVocabList(), 
					  fTrainVcbList.getVocabList(), 
					  m1.getETotalWCount(), 
					  m1.getFTotalWCount(), true);
     
     string perp_filename = Prefix + ".perp" ;
     ofstream of_perp(perp_filename.c_str());
     
     cerr << "Writing PERPLEXITY report to: " << perp_filename << '\n';
     if(!of_perp){
       cerr << "\nCannot write to " << perp_filename <<'\n';
       exit(1);
     }

     if (testCorpus)
       generatePerplexityReport(trainPerp, testPerp, trainViterbiPerp, 
				testViterbiPerp, of_perp, corpus.getTotalNoPairs(), 
				(*testCorpus).getTotalNoPairs(), //lastmodel, 
				true);
     else 
       generatePerplexityReport(trainPerp, testPerp, trainViterbiPerp, testViterbiPerp, 
				of_perp, corpus.getTotalNoPairs(), 0, true);
     
     string ce_filename = Prefix + ".ce" ;
     ofstream of_ce(ce_filename.c_str());
     cerr << "Writing CROSS-ENTROPY report to: " << ce_filename << '\n';
     if(!of_ce){
       cerr << "\nCannot write to " << ce_filename <<'\n';
       exit(1);
     }
     if( testCorpus )
       generatePerplexityReport(trainPerp, testPerp, trainViterbiPerp, 
				testViterbiPerp, of_ce, corpus.getTotalNoPairs(), 
				(*testCorpus).getTotalNoPairs(), false);
     else 
       generatePerplexityReport(trainPerp, testPerp, trainViterbiPerp, testViterbiPerp, 
				of_ce, corpus.getTotalNoPairs(), 0, false);

     //print train vocabulary list 
     
     string eTrainVcbFile = Prefix + ".trn.src.vcb" ;
     ofstream of_eTrainVcb(eTrainVcbFile.c_str());
     cerr << "Writing source vocabulary list to : " << eTrainVcbFile << '\n';
     if(!of_eTrainVcb){
       cerr << "\nCannot write to " << eTrainVcbFile <<'\n';
       exit(1);
     }
     eTrainVcbList.printVocabList(of_eTrainVcb) ;
     
     string fTrainVcbFile = Prefix + ".trn.trg.vcb" ;
     ofstream of_fTrainVcb(fTrainVcbFile.c_str());
     cerr << "Writing source vocabulary list to : " << fTrainVcbFile << '\n';
     if(!of_fTrainVcb){
       cerr << "\nCannot write to " << fTrainVcbFile <<'\n';
       exit(1);
     }
     fTrainVcbList.printVocabList(of_fTrainVcb) ;

     //print test vocabulary list 
     
     string eTestVcbFile = Prefix + ".tst.src.vcb" ;
     ofstream of_eTestVcb(eTestVcbFile.c_str());
     cerr << "Writing source vocabulary list to : " << eTestVcbFile << '\n';
     if(!of_eTestVcb){
       cerr << "\nCannot write to " << eTestVcbFile <<'\n';
       exit(1);
     }
     eTestVcbList.printVocabList(of_eTestVcb) ;
     
     string fTestVcbFile = Prefix + ".tst.trg.vcb" ;
     ofstream of_fTestVcb(fTestVcbFile.c_str());
     cerr << "Writing source vocabulary list to : " << fTestVcbFile << '\n';
     if(!of_fTestVcb){
       cerr << "\nCannot write to " << fTestVcbFile <<'\n';
       exit(1);
     }
     fTestVcbList.printVocabList(of_fTestVcb) ;

     if (testCorpus)
       printOverlapReport(m1.getTTable(), corpus, *testCorpus, eTrainVcbList, 
			  fTrainVcbList, eTestVcbList, fTestVcbList);
   } // end of if(!NODUMPS)

   fn = time(NULL);    // finish time
   cerr << '\n' << "Entire Training took: " << difftime(fn, st1) << " seconds\n";
   cerr << "Program Finished at: "<< ctime(&fn) << '\n';
   cerr << "==========================================================\n";
   return 0;
}
