/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
#ifndef _model1_h
#define _model1_h 1

#include <assert.h>
 
#include <iostream>
#include <strstream>
#include <algorithm>
#include <functional>
#include <map>
#include <set>
#include <utility>
#include <hash_map>
#include <time.h>
#include <fstream.h>
#include <math.h>
#include <stdio.h>

#include "Vector.h"
#include "vocab.h"
#include "TTables.h"
#include "getSentence.h"
#include "Perplexity.h"
#include "vocab.h"
#include "Dictionary.h"

class model1{
private:
  string efFilename;
  vcbList&  Elist ;
  vcbList&  Flist ;
  float eTotalWCount ; // size of source copus in number of words
  float fTotalWCount ; // size of target corpus in number of words 
  int noEnglishWords;
  int noFrenchWords;
  tmodel<COUNT, PROB> tTable;
  Vector<WordEntry>& evlist ;
  Vector<WordEntry>& fvlist ;
  //  tmodel<float, LogProb> tCountTable;
  //  string eFilename;
  //  string fFilename;
public:
  //  model1(const char* ename, const char* fname, const char* efname);
  model1 (const char* efname, vcbList& evcblist, vcbList& fvcblist);

  // methods
  void initialize_table_uniformly(sentenceHandler& sHandler1);
  void em_with_tricks(int noIterations, Perplexity& perp, sentenceHandler& sHandler1 , bool seedModel1, Dictionary& dictionary, bool useDict, 
		      Perplexity*, sentenceHandler*, Perplexity&, Perplexity *);

  void load_table(const char* tname);
  void readVocabFile(const char* fname, Vector<WordEntry>& vlist, int& vsize, 
		     int& total);

  // access functions
  inline const Vector<WordEntry>& getEnglishVocabList(void)const {return Elist.getVocabList();};
  inline const Vector<WordEntry>& getFrenchVocabList(void)const  {return Flist.getVocabList();};
  inline const float getETotalWCount(void) const {return eTotalWCount;};
  inline const float getFTotalWCount(void) const {return fTotalWCount;};
  inline const int getNoEnglishWords(void) const  {return noEnglishWords;};
  inline const int getNoFrenchWords(void)  const {return noFrenchWords;};
  inline tmodel<COUNT, PROB>& getTTable(void) {return tTable;};
  //  inline tmodel<float, LogProb>& getTCountTable(void) {return tCountTable;};
  //  inline string& getEFilename(void) {return eFilename;};
  //  inline string& getFFilename(void) {return fFilename;};
  inline string& getEFFilename(void) {return efFilename;};
  //inline void clear(void) {tTable.clear();};
 private:
  void em_loop(int it, Perplexity& perp, sentenceHandler& sHandler1, bool seedModel1, bool , const char*, Dictionary& dictionary, bool useDict, 
	       Perplexity& viterbiperp, bool=false);
  friend class model2;
};



#endif
