/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
#include "model2.h"
#include "Globals.h"
#include "utility.h"

model2::model2(model1& m): Elist(m.Elist), Flist(m.Flist), aTable(false, false, false, false), aCountTable(false, ADTableIgnoreL, ADTableIgnoreM, ADTableSpecial) , efFilename(m.getEFFilename())
, eTotalWCount(m.getETotalWCount()), fTotalWCount(m.getFTotalWCount())
{  }

void model2::initialize_table_uniformly(){
  // initialize the aTable uniformly (run this before running em_with_tricks)
  for(WordIndex l=1; l < MAX_SENTENCE_LENGTH; l++){
    PROB uniform_val = 1.0 / (l+1) ;
    for(WordIndex m=1; m < MAX_SENTENCE_LENGTH; m++)
      for(WordIndex j=1; j <= m; j++)
	for(WordIndex i=0; i <= l; i++)
      	  aTable.setValue(i, j, l, m, uniform_val);
  }
  
}



void model2::em_with_tricks(tmodel<COUNT, PROB>& tTable, int noIterations, Perplexity& perp, 
			    sentenceHandler& sHandler1,  Perplexity* testPerp, 
			    sentenceHandler* testHandler, Perplexity& trainViterbiPerp, 
			    Perplexity& testViterbiPerp)
{

  time_t it_st, st, it_fn, fn;
  string tfile, afile, tfile_actual, tcfile, number, alignfile, test_alignfile;
  int pair_no = 0;
  bool dump_files = false ;
  ofstream of2 ;
  //  Perplexity perp ;
  st = time(NULL) ;
  //  sentenceHandler sHandler1(efFilename.c_str());
  sHandler1.rewind();
  cerr << "\n==========================================================\n";
  cerr << "Model2 Training Started at: " << ctime(&st) << "\n";
  for(int it=1; it <= noIterations ; it++){
    pair_no = 0;
    it_st = time(NULL) ;
    cerr << endl << "-----------\nModel2: Iteration " << it << '\n';
    dump_files = (Model2_Dump_Freq != 0) && ((it % Model2_Dump_Freq) == 0) && !NODUMPS;
    if (dump_files){
      // set up the names of the files where the tables will be printed 
      number = "";
      int n = it;
      do{
	//mj changed next line
      number.insert((unsigned int)0, 1, (char)(n % 10 + '0'));
      } while((n /= 10) > 0);
      tfile = Prefix + ".t2." + number ;
      tfile_actual = Prefix + ".actual.t2." + number ;
      tcfile = Prefix + ".tc2." + number ;
      afile = Prefix + ".a2." + number ;
      alignfile = Prefix + ".A2." + number ;
      test_alignfile = Prefix + ".tst.A2." + number ;
    }
    // clear tCountTable, aCountTable
    //    tCountTable.clear();
    aCountTable.clear();
    // collect counts for training corpus
    em_loop(1, tTable, it, perp, sHandler1, dump_files, alignfile.c_str(), trainViterbiPerp, false);
    if (testPerp && testHandler)
      em_loop(1, tTable, it, *testPerp, *testHandler, dump_files, test_alignfile.c_str(), testViterbiPerp, 
	      true); 
    
    if (dump_files){
      of2.close(); // close alignments file 
      //      tTable.printCountTable(tcfile.c_str(), evlist, fvlist, true);
    }
    cerr << endl << "Model2 Iteration " << it << " done. I normalize now...";
    
    // normalize tc 
    tTable.normalizeTable(Elist, Flist);
    aCountTable.normalize(aTable);
    cerr << "\nModel2: ("<<it<<") TRAIN CROSS-ENTROPY " << perp.cross_entropy()
	 << " PERPLEXITY " << perp.perplexity() << '\n';
     if (testPerp && testHandler)
       cerr << "\nModel2: ("<<it<<") TEST CROSS-ENTROPY " << (*testPerp).cross_entropy()
	    << " PERPLEXITY " << (*testPerp).perplexity() 
	    << '\n';
     cerr << "\nModel2: ("<<it<<") VITERBI TRAIN CROSS-ENTROPY " << trainViterbiPerp.cross_entropy()
	 << " PERPLEXITY " << trainViterbiPerp.perplexity() << '\n';
    if (testPerp && testHandler)
       cerr << "\nModel2: ("<<it<<") VITERBI TEST CROSS-ENTROPY " << testViterbiPerp.cross_entropy()
	    << " PERPLEXITY " << testViterbiPerp.perplexity() 
	    << '\n';

    cerr << "tTable contains " << tTable.getHash().bucket_count() 
	 << " buckets and  " << tTable.getHash().size() << " entries." ;
    
    if (dump_files){
      // print the tTable, aTable
      tTable.printProbTable(tfile.c_str());
      //      tTable.printTableWithActualVocab(tfile_actual.c_str(), evlist, fvlist);
      aTable.printTable(afile.c_str());
    }
    it_fn = time(NULL) ;
    cerr << "\nModel 2 Iteration: " << it<< " took: " <<
      difftime(it_fn, it_st) << " seconds\n";
  } // end of iterations 
  aCountTable.clear();
  fn = time(NULL) ;
  cerr << "\nModel2 Training Finished at: " << ctime(&fn) << endl;
  cerr << endl << "Entire Model2 Training took: " << difftime(fn, st) << " seconds\n";
  cerr << "==========================================================\n";

}

void model2::load_table(const char* aname){
  /* This function loads the a table from the given file; use it
     when you want to load results from previous a training without
     doing any new training.
     NAS, 7/11/99
  */
  cerr << "Model2: loading a table \n";
  aTable.readTable(aname);
}


void model2::em_loop(int model, tmodel<COUNT, PROB>& tTable, 
		     int it, Perplexity& perp, sentenceHandler& sHandler1, 
		     bool dump_alignment, const char* alignfile, Perplexity& viterbi_perp, 
		     bool test)
{
  WordIndex i, j, l, m ;
  double cross_entropy;
  int pair_no=0 ;
  perp.clear();
  viterbi_perp.clear();
  ofstream of2;
  // for each sentence pair in the corpus
  if (dump_alignment)
    of2.open(alignfile);
  sentPair sent ;
  sHandler1.rewind();
  while(sHandler1.getNextSentence(sent)){
    Vector<WordIndex>& es = sent.eSent;
    Vector<WordIndex>& fs = sent.fSent;
    const float so  = sent.noOccurrences;
    if ((sent.sentenceNo % 10000 ) == 0)
	cerr << sent.sentenceNo << '\n';
    l = es.size() - 1;
    m = fs.size() - 1;
    cross_entropy = log(1);
    Vector<WordIndex> viterbi_alignment(fs.size());
    double viterbi_score = 1;
    for(j=1; j <= m; j++){
      Vector<LpPair<COUNT,PROB> *> sPtrCache(es.size(),0); // cache pointers to table 
      // entries  that map fs to all possible ei in this sentence.
      PROB denom = 0.0;
      PROB e = 0.0, word_best_score = 0;
      WordIndex best_i = 0 ; // i for which fj is best maped to ei
      for(i=0; i <= l; i++){
	sPtrCache[i] = tTable.getPtr(es[i], fs[j]) ;
	if (sPtrCache[i] != 0 &&(*(sPtrCache[i])).prob > PROB_SMOOTH ) 
	  // if valid pointer
	  e = (*(sPtrCache[i])).prob * aTable.getValue(i, j, l, m) ;
	else e = PROB_SMOOTH * aTable.getValue(i, j, l, m);
	denom += e ;
	if (e > word_best_score){
	  word_best_score = e ;
	  best_i = i ;
	}
      }
      viterbi_alignment[j] = best_i ;
      viterbi_score *= word_best_score; ///denom ;
      cross_entropy += log(denom) ;
      if (denom == 0){
	if (test)
	  cerr << "WARNING: denom is zero (TEST)\n";
	else 
	  cerr << "WARNING: denom is zero (TRAIN)\n";
      }      
      if (!test){
	if(denom > 0){	  
	  COUNT val = COUNT(so) / (COUNT) double(denom) ;
	  for( i=0; i <= l; i++){
	    PROB e(0.0);
	    if (sPtrCache[i] != 0 &&  (*(sPtrCache[i])).prob > PROB_SMOOTH)
	      e = (*(sPtrCache[i])).prob ;
	    else e = PROB_SMOOTH  ;
	    e *= aTable.getValue(i, j, l, m);
	    COUNT temp = COUNT(e) * val ;
	    if (sPtrCache[i] != 0) 
	      (*(sPtrCache[i])).count += temp ;
	    else 	      
	      tTable.incCount(es[i], fs[j], temp);	    
	    aCountTable.getRef(i, j, l, m)+= temp ;
	  } /* end of for i */
	} // end of if (denom > 0)
      }// if (!test)
    } // end of for (j) ;
    perp.addFactor(cross_entropy, so, l, m);
    viterbi_perp.addFactor(log(viterbi_score), so, l, m);
    //    cerr << "P(V, f/e) = " << viterbi_score << " P(f/e) = " << exp(cross_entropy) << '\n';
    if (dump_alignment)
      printAlignToFile(es, fs, Elist.getVocabList(), Flist.getVocabList(), of2, viterbi_alignment, sent.sentenceNo, viterbi_score);
    pair_no++;
  } /* of while */
  cerr << "\nProcessed a total of: "<<pair_no<<" sentence pairs.\n"; 
  sHandler1.rewind();
  perp.record(model, it);
  viterbi_perp.record(model, it);
}
