/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
#include "model3.h"
#include "collCounts.h"
#include "Globals.h"
#include "utility.h"

#define TRICKY_IBM3_TRAINING


/*double factorial(int n)
  {
double  f=1;
for(int i=2; i <= n; i++)
  f *= i;
return f;
}
*/


model3::model3(model1& m1, model2& m2): /*noEnglishWords(m1.getNoEnglishWords()), 
					  noFrenchWords(m1.getNoFrenchWords()), */
  Elist(m2.Elist), Flist(m2.Flist), 
  efFilename(m1.getEFFilename()), /*evlist(m1.getEnglishVocabList()), fvlist(m1.getFrenchVocabList()), */eTotalWCount(m1.getETotalWCount()), fTotalWCount(m1.getFTotalWCount()), dTable(true, false, false, false), dCountTable(true, ADTableIgnoreL, ADTableIgnoreM, ADTableSpecial), nTable(m1.getNoEnglishWords()+1, MAX_FERTILITY), nCountTable(m1.getNoEnglishWords()+1, MAX_FERTILITY)
{}
      


void model3::load_tables(const char *nfile, const char *dfile, const char *p0file){
  cerr << "Model3: loading n, d, p0 tables \n";

  nTable.readNTable(nfile);
  dTable.readTable(dfile);
  ifstream inf(p0file);
  cerr << "Reading p0 value from " << p0file << "\n";
  inf >> p0;
  inf.close();
  p1 = 1 - p0;
}
  
model3::~model3()
{
  dTable.clear();
  dCountTable.clear();
  nTable.clear();
  nCountTable.clear();
}


void model3::em(tmodel<COUNT, PROB>& tTable, 
		amodel<PROB>& aTable, 
		amodel<COUNT>& aCountTable, 
		int noIterations, sentenceHandler& sHandler1)
{

  double p0_count, p1_count ;
  LogProb all_prob, aprob, temp ;
  WordIndex i, j, l, m ;
  time_t it_st, st, it_fn, fn ;
  string tfile, dfile, nfile, p0file, afile, number;
  
  st = time(NULL) ;
  if (Log)
    logmsg << "\n" << "Starting Model3:  Training";
  cerr << "\n" << "Starting Model3:  Training";
  //  sentenceHandler sHandler1(efFilename.c_str());
  sHandler1.rewind();
  for(int it=1; it <= noIterations; it++){
    it_st = time(NULL) ;
    if (Log)
      logmsg << "\n" << "Model3: Iteration " << it;
    cerr << "\n" << "Model3: Iteration " << it;
    
    // set up the names of the files where the tables will be printed 
    int n = it;
    number = "";
    do{
      //mj changed next line
      number.insert((unsigned int) 0, 1, (char)(n % 10 + '0'));
    } while((n /= 10) > 0);
    tfile = Prefix + ".t3." + number ;
    afile = Prefix + ".a3." + number ;
    nfile = Prefix + ".n3." + number ;
    dfile = Prefix + ".d3." + number ;
    p0file = Prefix + ".p0_3." + number ;
    //    tCountTable.clear();
    dCountTable.clear();
    nCountTable.clear();
    p0_count = p1_count = 0 ;
    all_prob = 0 ;
    sentPair sent ;
    while(sHandler1.getNextSentence(sent)){
      Vector<WordIndex>& es = sent.eSent;
      Vector<WordIndex>& fs = sent.fSent;
      const float count  = sent.noOccurrences;
      if ((sent.sentenceNo % 1000) == 0)
      cerr <<sent.sentenceNo << '\n'; 
      Vector<WordIndex> A(fs.size(),-1);
      Vector<WordIndex> Fert(es.size(),0);
      LogProb lcount=(LogProb)count;
      l = es.size()-1;
      m = fs.size()-1;
      WordIndex x, y ;
      all_prob = prob_of_target_given_source(tTable, fs, es);
      if (all_prob == 0)
	cerr << "\n" <<"all_prob = 0"; 
	
      for ( x = 0 ; x < pow(l+1, m) ; x++){ // For all possible alignmets A
	y = x ;
	for (j = 1 ; j <= m ; j++){
	  A[j] = y % (l+1) ;
	  y /= (l+1) ;
	}
	for(i = 0 ; i <= l ; i++)
	  Fert[i] = 0 ;
	for (j = 1 ; j <= m ; j++)
	  Fert[A[j]]++;
	if (2 * Fert[0] <= m){ /* consider alignments that has Fert[0] less than
				 half the number of words in French sentence */
	  aprob = prob_of_target_and_alignment_given_source(A, Fert, tTable, fs, es);
	  temp = aprob/all_prob ;
	  LogProb templcount = temp*lcount;

	  for (j = 1 ; j <= m ; j++){
	    tTable.incCount(es[A[j]], fs[j], templcount);
	    if (0 != A[j])
	      dCountTable.getRef(j, A[j], l, m)+=templcount;
	  }
	  for(i = 0 ; i <= l ; i++)
	    {
	      nCountTable.getRef(es[i], Fert[i])+=templcount;
	      //cerr << "AFTER INC2: " << templcount << " " << nCountTable.getRef(es[i], Fert[i]) << endl;
	    }
	  p1_count +=  double(temp) * (Fert[0] * count) ;
	  p0_count +=  double(temp) * ((m - 2 * Fert[0]) * count) ;
	}
      } /* of looping over all alignments */
    } /* of sentence pair E, F */
    sHandler1.rewind();
    
    // normalize tables

    cerr << "I normalize now ..." ;
    tTable.normalizeTable(Elist, Flist);
    aCountTable.normalize(aTable);
    dCountTable.normalize(dTable);
    nCountTable.normalize(nTable);

    // normalize p1 & p0 
    
    if (p1_count + p0_count != 0){
      p1 = p1_count / ( p1_count + p0_count ) ;
      p0 = 1 - p1 ;
    }
    else {
      p1 = p0 = 0 ;
    }
    cerr << "Done. " ;
    
    // print tables 
    tTable.printProbTable(tfile.c_str());
    dTable.printTable(dfile.c_str());
    nTable.printNTable(Elist.uniqTokens(), nfile.c_str(), Elist.getVocabList(), false);
    ofstream of(p0file.c_str());
    of << p0;
    of.close();
    it_fn = time(NULL) ;
    cerr << "\n" << "Model3 Iteration "<<it<<" took: " << difftime(it_fn, it_st) << " seconds\n";
    
  } /* of iterations */
  fn = time(NULL) ;
  cerr << "\n" << "Entire Model3 Training took: " << difftime(fn, st) << " seconds\n";
}



      
	  
	
	
//-----------------------------------------------------------------------





void model3::viterbi(tmodel<COUNT, PROB>& tTable, 
		     amodel<PROB>& aTable, 
		     amodel<COUNT>& aCountTable, 
		     int noIterationsModel3, int noIterationsModel4Prime, Perplexity& perp, Perplexity& trainVPerp, 
		     sentenceHandler& sHandler1, Perplexity* testPerp, Perplexity* testVPerp, 
		     sentenceHandler* testHandler)
{
  int noIterations=noIterationsModel3+noIterationsModel4Prime;
  int itTransfer=noIterationsModel3+1;
  d4model d4m(MAX_SENTENCE_LENGTH);
  double p0_count, p1_count ;
  time_t it_st, st, it_fn, fn;
  bool dump_files = false ;
  unsigned int model = 0 ;
  string tfile, tfile_actual, dfile, afile, nfile, nfile_actual, p0file, alignfile, number, test_alignfile, d4file;
  st = time(NULL);
  sHandler1.rewind();
  if (testPerp && testHandler)
    (*testHandler).rewind();
  string fullModelName;
  if( noIterations<itTransfer )
    fullModelName="Model3";
  else
    fullModelName="Model3+4";
  
  cerr << "\n==========================================================\n";
  cerr << "Starting "<<fullModelName<<":  Viterbi Training";
  if (Log){
    logmsg << "\n==========================================================\n";
    logmsg << "Starting "<<fullModelName<<":  Viterbi Training";
  }
  cerr << "\n "<<fullModelName<<" Training Started at: "<< ctime(&st) << '\n';
  for(int it=1; it <= noIterations; it++){
    string modelName;
    if( it<itTransfer ){
      model = 3 ; // means model 3.
      modelName="Model3";
    }
    else if( it==itTransfer ){
      model = 4 ; // means model 3 to 4 transfer 
      modelName="Transfer3To4'";
    }
    else {
      model = 5 ;
      modelName="Model4'";
    }
    it_st = time(NULL);
    cerr <<"\n---------------------\n"<<modelName<<": Iteration " << it<<'\n';
    if (Log) 
      logmsg <<"\n---------------------\n"<<modelName<<": Iteration " << it<<'\n';
    dump_files = (Model3_Dump_Freq != 0) && ((it % Model3_Dump_Freq) == 0) && !NODUMPS ;
    if (dump_files){
      // set up the names of the files where the tables will be printed 
      int n = it;
      number = "";
      do{
	//mj changed next line
	number.insert((unsigned int) 0, 1, (char)(n % 10 + '0'));
      } while((n /= 10) > 0);
      tfile = Prefix + ".t3." + number ;
      tfile_actual = Prefix + ".actual.t3." + number ;
      afile = Prefix + ".a3." + number ;
      nfile = Prefix + ".n3." + number ;
      nfile_actual = Prefix + ".actual.n3." + number ;
      dfile = Prefix + ".d3." + number ;
      d4file = Prefix + ".d4." + number ;
      alignfile = Prefix + ".A3." + number ;
      test_alignfile = Prefix + ".tst.A3." + number ;
      p0file = Prefix + ".p0_3." + number ;
    }
    // clear count tables 
    //    tCountTable.clear();
    dCountTable.clear();
    aCountTable.clear();
    nCountTable.clear();
    p0_count = p1_count = 0 ;

#ifdef TRICKY_IBM3_TRAINING
    if( it<itTransfer )
      {
	viterbi_loop_with_tricks<transpair_model3, d4model>(tTable, aTable, aCountTable, it, perp, 
							   trainVPerp, sHandler1, 
							   dump_files, alignfile.c_str(), 
							   p0_count, p1_count, 
							   true, model, (d4model*)0);
	if (testPerp && testHandler)
	  viterbi_loop_with_tricks<transpair_model3, d4model>(tTable, aTable, aCountTable, it, 
							     *testPerp, *testVPerp, 
							     *testHandler, dump_files, 
							     test_alignfile.c_str(), 
							     p0_count, p1_count, false, model, 
							     (d4model*)0);
      }
    else if(it==itTransfer)
      {
	
	cerr<<"TRANSFER FROM MODEL 3 TO MODEL 4': \n";
	viterbi_loop_with_tricks<transpair_model3, d4model>(tTable, aTable, aCountTable, it, perp, trainVPerp, sHandler1, 
				 dump_files, alignfile.c_str(), p0_count, p1_count, 
				 true, model, &d4m);
	if (testPerp && testHandler)
	  viterbi_loop_with_tricks<transpair_model3, d4model>(tTable, aTable, aCountTable, it, *testPerp, *testVPerp, 
				   *testHandler, dump_files, test_alignfile.c_str(), 
				   p0_count, p1_count, false, model, &d4m);	
	d4m.normalizeTable();
	ofstream d4output(d4file.c_str());
	d4output << d4m;
      }
    else if(it>itTransfer)
      {
	cerr<<"TRAINING OF MODEL 4': \n";
	viterbi_loop_with_tricks<transpair_model4, d4model>(tTable, aTable, aCountTable, it, perp, trainVPerp, sHandler1, 
				dump_files, alignfile.c_str(), p0_count, p1_count, 
				true, model, &d4m);
	if (testPerp && testHandler)
	  viterbi_loop_with_tricks<transpair_model4, d4model>(tTable, aTable, aCountTable, it, *testPerp, *testVPerp, 
				  *testHandler, dump_files, test_alignfile.c_str(), 
				  p0_count, p1_count, false, model, &d4m);	
	d4m.normalizeTable();
      }
    
#else
    viterbi_loop(tTable, aTable, aCountTable, it, perp, trainVPerp, sHandler1, dump_files, 
		 alignfile.c_str(), p0_count, p1_count, true, model);
    if (testPerp && testHandler)
      viterbi_loop(tTable, aTable, aCountTable, it, *testPerp, *testVPerp, *testHandler1, 
		   dump_files, test_alignfile.c_str(), p0_count, p1_count, false, model);
 
#endif		 

    // now normalize count tables 
    cerr << "I normalize now ..." ;
        
    tTable.normalizeTable(Elist, Flist);
    aCountTable.normalize(aTable);
    dCountTable.normalize(dTable);
    nCountTable.normalize(nTable);

    cerr << "tTable contains " << 
      tTable.getHash().bucket_count() << " buckets and "<<
      tTable.getHash().size() << " entries.\n";

    // normalize p1 & p0 
    
    
    if (FIX_P0){
      p0 = P0 ;
      p1 = 1 - P0 ;
    }
    else {
      if (p1_count + p0_count != 0){
	p1 = p1_count / ( p1_count + p0_count ) ;
	p0 = 1 - p1 ;
      }
      else {
	p1 = p0 = 0 ;
      }
    }
    
    cerr << modelName<<":("<<it<<") TRAIN CROSS-ENTROPY " << perp.cross_entropy()
	 << " PERPLEXITY " << perp.perplexity() << '\n';
    if (testPerp && testHandler)
      cerr << modelName << ":("<<it<<") TEST CROSS-ENTROPY " << (*testPerp).cross_entropy()
	   << " PERPLEXITY " << (*testPerp).perplexity() << '\n';
    //    " sum: " << (*testPerp).getSum()<< " wc: " << (*testPerp).word_count() << '\n';
    cerr << "Model3: ("<<it<<") TRAIN VITERBI CROSS-ENTROPY " << trainVPerp.cross_entropy()
	 << " PERPLEXITY " << trainVPerp.perplexity() << '\n';
    if (testPerp && testHandler)
      cerr << "Model3:  ("<<it<<")TEST VITERBI CROSS-ENTROPY " << (*testVPerp).cross_entropy()
	   << " PERPLEXITY " << (*testVPerp).perplexity() << '\n';
    //" Sum: " << (*testVPerp).getSum() << " wc: " << (*testVPerp).word_count() << '\n';
    if (dump_files){
      // print tables 
      tTable.printProbTable(tfile.c_str());
      tTable.printProbTableWithActualVocab(tfile_actual.c_str(), Elist.getVocabList(), Flist.getVocabList());
      aTable.printTable(afile.c_str());
      dTable.printTable(dfile.c_str());
      nTable.printNTable(Elist.uniqTokens(), nfile.c_str(), Elist.getVocabList(), false);
      nTable.printNTable(Elist.uniqTokens(), nfile_actual.c_str(), Elist.getVocabList(), true);
      ofstream of(p0file.c_str());
      of << p0;
      of.close();
      cerr << "Done."  ;
    }
    
    it_fn = time(NULL) ;
    cerr << "\n" << modelName << " Viterbi Iteration : "<<it<<  " took: " <<
      difftime(it_fn, it_st) << " seconds\n";
  } /* of iterations */
  fn = time(NULL);
  cerr << fullModelName<<" Training Finished at: " << ctime(&fn) << "\n";
  cerr << "\n" << "Entire Viterbi "<<fullModelName<<" Training took: " << difftime(fn, st) << " seconds\n";
  cerr << "==========================================================\n";
}




