/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
#ifndef _model3_h
#define _model3_h 1

#include <assert.h>
 
#include <iostream>
#include <algorithm>
#include <functional>
#include <map>
#include <set>
//#include <vector>
#include "Vector.h"
#include <utility>
#include <hash_map>
#include <time.h>
#include <fstream.h>
#include <math.h>
#include "MoveSwapMatrix.h"
#include "TTables.h"
#include "ATables.h"
#include "NTables.h"
#include "AlignTables.h"
#include "getSentence.h"
#include "defs.h"
#include "model2.h"
#include "Perplexity.h"
#include "transpair_model3.h"
#include "alignment.h"
#include "vocab.h"
#include "D4Tables.h"

class model3{
private:
  //  typedef float PROB;
  //  typedef float COUNT;

  const vcbList&Elist;
  const vcbList&Flist;
  //int noEnglishWords;
  //int noFrenchWords;

  //  string eFilename;
  //  string fFilename;
  string efFilename;

  //const Vector<WordEntry>& evlist ;
  //const Vector<WordEntry>&  fvlist;

  float eTotalWCount ; // size of source copus in number of words
  float fTotalWCount ; // size of target corpus in number of words 

  amodel<PROB>   dTable;
  amodel<COUNT>  dCountTable;

  PROB p0;
  PROB p1;

  nmodel<PROB>  nTable;
  nmodel<COUNT> nCountTable;

public:
  model3(model1& m1, model2& m2);
  ~model3();
  // methods
  void transfer(model1& m1, model2 &m2, sentenceHandler&, bool, Perplexity&, Perplexity&);
  void transferSimple(model1& m1, model2& m2, sentenceHandler&, bool, Perplexity&, Perplexity&);
  void load_tables(const char *nfile, const char *dfile, const char *p0file);
  
//  void printTables(model1& m1, model2& m2);
  void em(tmodel<COUNT, PROB>&, amodel<PROB>&, amodel<COUNT>&, int, sentenceHandler&);
  void viterbi(tmodel<COUNT, PROB>&, amodel<PROB>&, amodel<COUNT>&, 
	       int, int, Perplexity&, Perplexity&, sentenceHandler&, Perplexity*, Perplexity*, sentenceHandler*);
    
private:
  LogProb prob_of_special(Vector<WordIndex>&, 
			  Vector<WordIndex>&, 
			  tmodel<COUNT, PROB>&, 
			  Vector<WordIndex>&, 
			  Vector<WordIndex>&);

  LogProb prob_of_target_and_alignment_given_source(Vector<WordIndex>&, 
						  Vector<WordIndex>&, 
						  tmodel<COUNT, PROB>&, 
						  Vector<WordIndex>&, 
						  Vector<WordIndex>&);
  LogProb prob_of_target_given_source(tmodel<COUNT, PROB>&, 
				    Vector<WordIndex>&, 
				    Vector<WordIndex>&);

  LogProb scoreOfMove(Vector<WordIndex>&, Vector<WordIndex>&, 
		    Vector<WordIndex>&, Vector<WordIndex>&, 
		    tmodel<COUNT, PROB>&, WordIndex, WordIndex);

  LogProb scoreOfSwap(Vector<WordIndex>&, Vector<WordIndex>&, 
		      Vector<WordIndex>&, tmodel<COUNT, PROB>&, int, int);
  
  void hillClimb(Vector<WordIndex>&, Vector<WordIndex>&, 
		 Vector<WordIndex>&, Vector<WordIndex>&, 
		 LogProb&, tmodel<COUNT, PROB>&, int, int);

  void findBestAlignment(Vector<WordIndex>&, Vector<WordIndex>&, 
			 Vector<WordIndex>&, Vector<WordIndex>&, 
			 LogProb&, tmodel<COUNT, PROB>&, amodel<PROB>&, int , int);


  void findAlignmentsNeighborhood( tmodel<COUNT, PROB>&,  amodel<PROB>&, double&, 
				   double&, 
				   Vector<WordIndex>&, 
				   Vector<WordIndex>&, 
				   LogProb&align_total_count, 
				   alignmodel&neighborhood, 
				   int, int);
  void collectCountsOverAlignement(tmodel<COUNT, PROB>&, amodel<COUNT>&, 
				   const Vector<WordIndex>& es, 
				   const Vector<WordIndex>& fs, 
				   double&, double&, 
				   const Vector<WordIndex>&, 
				   LogProb , int count);
  LogProb viterbi_model2(const transpair_model3&ef, alignment&output, int i_peg = -1 , int j_peg = -1 )const;
 private:
  void estimate_t_a_d(tmodel<COUNT, PROB>&, amodel<COUNT>& aCountTable, 
		      amodel<PROB>& aTable, 
		      sentenceHandler& sHandler1, Perplexity& perp, Perplexity& perp, 
		      bool simple, bool dump_files);
  void viterbi_loop(tmodel<COUNT, PROB>& , amodel<PROB>&, amodel<COUNT>& , int, 
		    Perplexity&, Perplexity&, sentenceHandler&, bool, const char*, double&, 
		    double&, bool, unsigned int model);

  template<class MODEL_TYPE, class A>
  void viterbi_loop_with_tricks(tmodel<COUNT, PROB>& , amodel<PROB>&, 
				amodel<COUNT>& , int, Perplexity&, Perplexity&, sentenceHandler&, 
				bool, const char*, double&, double&, bool, unsigned int model, A*d4m=0);
};



#endif
