/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
#include "model3.h"
#include "collCounts.h"
#include "utility.h"
#include "Globals.h"

bool UseLinkCache=0;    /// optimization for pegging

LogProb model3::viterbi_model2(const transpair_model3&ef, alignment&output, int i_peg = -1 , int j_peg = -1 )const
{
  WordIndex best_i=0;
  LogProb ss=1;
  WordIndex l = ef.get_l(), m=ef.get_m();
  Vector<WordIndex> Fert(l+1, (WordIndex)0);
  if ((j_peg != -1) && (i_peg != -1))
    {
      output.set(j_peg, i_peg);
      ss *= ef.get_t(i_peg, j_peg) * ef.get_a(i_peg, j_peg);
      if( ss==0 )
	cerr << "WARNING: already starting is zero: " << ef.get_t(i_peg, j_peg) << " " << ef.get_a(i_peg, j_peg) << endl;
    }
  else
    ss=1;
  for (WordIndex j = 1 ; j <= m ; j++)if (int(j) != j_peg)
    {
      LogProb score = 0 ;
      for (WordIndex i = 0 ; i <= l ; i++)
	{
	  if( Fert[i]+1<MAX_FERTILITY && (i != 0 ||  m>=(2 * (Fert[0] + 1))))
	    {
	      LogProb temp = ef.get_t(i, j) * ef.get_a(i, j);
	      if (temp > score )
		{
		  best_i = i ;
		  score = temp ;
		}
	    } 
	}
      if (score == 0){
	cerr << "WARNING: In searching for model2 best alignment\n";
	cerr << "Nothing was set for target token at position j: " << j << "\n";
	for (WordIndex i = 0 ; i <= l ; i++){
	  cerr << "i: " << i << "ttable("<<i<<", "<<j<<") = " <<
	    ef.get_t(i, j) << " atable(" << i<<", "<<j<<", "<<
	    l<<", "<<m<<") = "<< ef.get_a(i, j) << " product " <<
	    ef.get_t(i, j) * ef.get_a(i, j) ;
	  if ((Fert[i]+1 < MAX_FERTILITY) && ((i == 0 &&  (m >= 2*(Fert[0]+1)))
					      || (i != 0)))
	    cerr <<"Passed fertility condition \n";
	  else 
	    cerr <<"Failed fertility condition \n";
	}	    
      }
      else
	{
	  output.set(j, best_i);
	  Fert[best_i]++;
	}
      ss *= score;
    } 
  if (ss <= 0){
    cerr << "WARNING: Model2 viterbi alignment has zero score.\n" ;
    cerr << "Here are the different elements that made this alignment probability zero \n";
    cerr << "Source length " << l << " target length " << m << '\n';
    LogProb gg=1 ; // for debugging only ..... 
    for (WordIndex j = 1 ; j <= m ; j++)if (int(j) != j_peg){
      LogProb score = 0 ;
      LogProb a = 0, t =0 ;
      for (WordIndex i = 0 ; i <= l ; i++){
	//	if( Debug_Fert[i]+1<MAX_FERTILITY && (i != 0 ||  m>=(2 * (Debug_Fert[0] + 1)))){
	  LogProb temp = ef.get_t(i, j) * ef.get_a(i, j);
	  if (temp > score ){
	    score = temp ;
	    best_i = i ;
	    a = ef.get_a(i, j);
	    t =  ef.get_t(i, j) ;
	  }
	  //      }
      }
      gg *= score ;
      //      cerr << "best: fs[" << j << "] "<< ef.getfSent()[j] <<"  : es[" << best_i << "] " <<
      //	ef.geteSent()[best_i] << " ,  a: " << ef.get_a(best_i, j) << " t: " << t << " score " << score << "  product : " << gg << " ss " <<
      //	ss << '\n';
    }
    //   for(WordIndex i = 0 ; i <= l ; i++)
    //    cerr << "Fert["<<ef.geteSent()[i]<<"] selected " << Fert[i] << '\n';
  }
  return 0.0;
}

template<class TRANSPAIR>
LogProb greedyClimb(MoveSwapMatrix<TRANSPAIR>&msc2, int i_peg = -1, int j_peg = -1)
{
  WordIndex l = msc2.get_l(), m=msc2.get_m();
  int changed=0;
  do 
    {
      changed=0;
      for (WordIndex j = 1 ; j <= m ; j++)if (int(j) != j_peg)
	{
	  WordIndex aj=msc2(j);
	  for (WordIndex j1 = j + 1 ; j1 <= m; j1++)if((aj != msc2(j1)) && (int(j1) != j_peg)&&msc2.cswap(j, j1) > 1.0)
	    msc2.doSwap(j, j1), changed=1;
	  for (WordIndex i = 0 ; i <= l ; i++)if(i != aj &&(i != 0 || (m >= 2 * (msc2.fert(0)+1)))  && msc2.fert(i)+1<MAX_FERTILITY && msc2.cmove(i, j)>1.0)
	    msc2.doMove(i, j), changed=1;
	}
    } while (changed);
  return msc2.get_ef().prob_of_target_and_alignment_given_source(msc2);
}

template<class TRANSPAIR>
LogProb hillClimb(MoveSwapMatrix<TRANSPAIR>&msc2, int i_peg = -1, int j_peg = -1)
{
  WordIndex l = msc2.get_l(), m=msc2.get_m();
  int changes=0;
  int best_change_type=-1, best_change_v1=-1, best_change_v2=-1;
  do 
    {
      LogProb best_change_so_far = 1.00001 ;  
      best_change_type=0;
      for (WordIndex j = 1 ; j <= m ; j++)if (int(j) != j_peg)
	{
	  WordIndex aj=msc2(j);
	  for (WordIndex j1 = j + 1 ; j1 <= m; j1++)if((aj != msc2(j1)) && (int(j1) != j_peg))
	    {
	      LogProb change = msc2.cswap(j, j1);
	      if (change > best_change_so_far)
		{
		  best_change_so_far = change ;
		  best_change_type=1;
		  best_change_v1=j;
		  best_change_v2=j1;
		} 
	    } 
	  for (WordIndex i = 0 ; i <= l ; i++)if(i != aj &&(i != 0 || (m >= 2 * (msc2.fert(0)+1))) && msc2.fert(i)+1<MAX_FERTILITY)
	    {
	      LogProb change = msc2.cmove(i, j);
	      if (change > best_change_so_far)
		{ 
		  best_change_so_far = change ;
		  best_change_type=2;
		  best_change_v1=j;
		  best_change_v2=i;
		} 
	    }
	}
      if (best_change_type==1)
	msc2.doSwap(best_change_v1, best_change_v2);
      if (best_change_type==2)
	msc2.doMove(best_change_v2, best_change_v1);
      changes++;
      if( changes>40 )
	cerr << "WARNING: already " << changes << " iterations in hillclimb: " << best_change_so_far << " " << best_change_type << " " << best_change_v1 << " " << best_change_v2 << endl;
      if(changes>60 )
	{
	  cerr << msc2 << endl;
	  break;
	}
    } while (best_change_type);
  return msc2.get_ef().prob_of_target_and_alignment_given_source(msc2);
}


template<class MODEL_TYPE, class ADDITIONAL_MODEL_DATA>
void model3::viterbi_loop_with_tricks(tmodel<COUNT, PROB>& tTable, 
				      amodel<PROB>& aTable, 
				      amodel<COUNT>& aCountTable, int it, 
				      Perplexity& perp, Perplexity& viterbiPerp, sentenceHandler& sHandler1, 
				      bool dump_files, const char* alignfile, 
				      double& p0_count, double& p1_count, 
				      bool collect_counts, unsigned int model, 
				      ADDITIONAL_MODEL_DATA*d4m)
{
  WordIndex i, j, l, m ;
  ofstream of2 ;
  int pair_no;
  
  if (dump_files )
    of2.open(alignfile);
  pair_no = 0 ; // sentence pair number 
  // for each sentence pair in the corpus
  perp.clear() ; // clears cross_entrop & perplexity 
  viterbiPerp.clear() ; // clears cross_entrop & perplexity 
  sentPair sent ;
  while(sHandler1.getNextSentence(sent)){
    Vector<WordIndex>& es = sent.eSent;
    Vector<WordIndex>& fs = sent.fSent;
    const float count  = sent.noOccurrences;    
    if ((sent.sentenceNo % 1000) == 0)
      cerr <<sent.sentenceNo << '\n'; 
    time_t sent_s = time(NULL) ;
    pair_no++ ;
    l = es.size() - 1 ;
    m = fs.size() - 1 ;
    if (Log){
      logmsg << "Processing sentence pair:\n\t";
      printSentencePair(es, fs, logmsg);
      for (i = 0 ; i <= l ; i++)
	  logmsg << Elist.getVocabList()[es[i]].word << " ";
	logmsg << "\n\t";
	for (j = 1 ; j <= m ; j++)
	  logmsg << Flist.getVocabList()[fs[j]].word << " ";
	logmsg << "\n";
      } 

      LogProb align_total_count=0;
      alignment viterbi2alignment(l, m);
      MODEL_TYPE ef(es, fs, tTable, aTable, dTable, nTable, p1, p0, d4m);
      viterbi_model2(ef, viterbi2alignment);
      Vector<pair<MoveSwapMatrix<MODEL_TYPE>*, LogProb> >setOfGoodCenters(1);
      set<alignment> alignments;
      MoveSwapMatrix<MODEL_TYPE> *best = (setOfGoodCenters[0].first  = new MoveSwapMatrix<MODEL_TYPE>(ef, viterbi2alignment));
      MoveSwapMatrix<MODEL_TYPE> _viterbi(*best), *viterbi=&_viterbi; // please, don't delete this line (FJO)
      setOfGoodCenters[0].second = std::hillClimb(*best);
      alignments.insert(*best); 
      if (setOfGoodCenters[0].second <= 0){
	cerr << "WARNING: Hill Climbing yielded a zero score viterbi alignment for the following pair:\n";
	cerr << *setOfGoodCenters[0].first ;
	printSentencePair(es, fs, cerr);      
	if(Log){
	  logmsg << "WARNING: Hill Climbing yielded a zero scorevite rbi alignment for the following pair:\n";
	  printSentencePair(es, fs, logmsg);
	}
      }
      if ( Peg )
	{
	  const MoveSwapMatrix<MODEL_TYPE> *useMatrix=viterbi;  // it is faster using 'best', ... (FJO)
	  Array2<short, vector<short> > linkCache(l+1, m+1, false);
	  if(UseLinkCache)for(unsigned int j=1;j<=m;j++)linkCache((*useMatrix)(j), j)=1;
	  for(WordIndex j=1;j<=m;j++)for(WordIndex i=0;i<=l;i++)
	    if( i!=(*useMatrix)(j) && (UseLinkCache==0||linkCache(i,j)==0) && 
		ef.get_t(i,j)>ef.get_t((*useMatrix)(j),j)*PEGGED_CUTOFF && 
		(i != 0 || (m >= 2 * (useMatrix->fert(0)+1))))
	    {
	      MoveSwapMatrix<MODEL_TYPE> *BESTPEGGED = new MoveSwapMatrix<MODEL_TYPE>(*useMatrix);
	      BESTPEGGED->doMove(i, j);
	      LogProb peggedAlignmentScore= std::hillClimb(*BESTPEGGED, i, j);
	      if( peggedAlignmentScore>setOfGoodCenters[0].second*(LogProb)PEGGED_CUTOFF && alignments.count(*BESTPEGGED)==0 )
		{
		  if( peggedAlignmentScore>setOfGoodCenters[0].second )
		    {
		      cout << "In sentence " << pair_no << " found a better alignment by pegging " << peggedAlignmentScore << " " << setOfGoodCenters[0].second << endl;
		      printAlignToFile(es, fs, Elist.getVocabList(), Flist.getVocabList(), cout, BESTPEGGED->getAlignment(), pair_no, setOfGoodCenters[0].second);
		    }
      		  setOfGoodCenters.push_back(make_pair(BESTPEGGED, peggedAlignmentScore));
		  alignments.insert(*BESTPEGGED);
		  assert( differences(*BESTPEGGED, *best)!=0 );
		  if(UseLinkCache)for(unsigned int j=1;j<=m;j++)linkCache((*BESTPEGGED)(j), j)=1;
		  BESTPEGGED=0;
		}
	      delete BESTPEGGED;
	    }
	} // end of if(Peg) 
      collectCountsOverNeighborhood(setOfGoodCenters, es, fs, tTable, aCountTable, dCountTable, nCountTable, p1_count, p0_count, align_total_count, count, collect_counts, d4m);
      perp.addFactor(log(double(align_total_count)), count, l, m);
      //      viterbiPerp.addFactor(log(double(setOfGoodCenters[0].second/align_total_count)), count, l, m);
      viterbiPerp.addFactor(log(double(setOfGoodCenters[0].second)), count, l, m);
      //      cerr << "P(V, f/e) = " << (setOfGoodCenters[0].second) << " P(f/e) = " 
      // << align_total_count << '\n';
      assert(log(double(setOfGoodCenters[0].second)) <= log(double(align_total_count)));
      if (dump_files)
	printAlignToFile(es, fs, Elist.getVocabList(), Flist.getVocabList(), of2, (setOfGoodCenters[0].first)->getAlignment(), pair_no, 
			 setOfGoodCenters[0].second);
      for(unsigned int i=0;i<setOfGoodCenters.size();i++)
	delete setOfGoodCenters[i].first;
      double period = difftime(time(NULL), sent_s);
      if (Log)
      	logmsg << "processing this sentence pair ("<<l+1<<"x"<<m<<") : "<<
      	  (l+1)*m << " took : " << period << " seconds\n";
      if (Verbose)
	cerr << "processing this sentence pair took : " << period
	     << " seconds\n";
      
    } /* of sentence pair E, F */
    cerr << "\nProcessed a total of: "<<pair_no<<" sentence pairs.\n"; 
    sHandler1.rewind();
    perp.record(model, it);
    viterbiPerp.record(model, it);
    if (dump_files)
      of2.close();

}

template
void model3::viterbi_loop_with_tricks<transpair_model3, d4model>(tmodel<COUNT, PROB>& tTable, 
								amodel<PROB>& aTable, 
								amodel<COUNT>& aCountTable, int it, 
								Perplexity& perp, Perplexity& viterbiPerp, sentenceHandler& sHandler1, 
								bool dump_files, const char* alignfile, 
								double& p0_count, double& p1_count, 
								bool collect_counts, unsigned int, 
								d4model*d4m);
template
void model3::viterbi_loop_with_tricks<transpair_model4, d4model>(tmodel<COUNT, PROB>& tTable, 
								amodel<PROB>& aTable, 
								amodel<COUNT>& aCountTable, int it, 
								Perplexity& perp, Perplexity& viterbiPerp, sentenceHandler& sHandler1, 
								bool dump_files, const char* alignfile, 
								double& p0_count, double& p1_count, 
								bool collect_counts, unsigned int, d4model*d4m);
