/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
/*--
transpair_model3: representation of a translation pair for model3 training
allowing for fast access (esp. to t table).

Franz Josef Och (30/07/99)
--*/
#ifndef transpair_model3_h_fjo_defined
#define transpair_model3_h_fjo_defined
#include "logprob.h"
#include "Array2.h"
#include "defs.h"
#include "Vector.h"
#include "NTables.h"
#include "ATables.h"
#include "TTables.h"
#include "alignment.h"
#include <math.h>

extern double factorial(int n);
inline bool doubleEqual(const double a, const double b)
{
  if( a==b )
    return 1.0;
  return fabs(1.0-a/b)<1e-10;
}


class transpair_model3 
{
 protected:
  Array2<PROB, Vector<PROB> > t, d, a, n;
  //Array2<short, Vector<short> > e;
  WordIndex l, m;
  PROB p0, p1;
 public:
  transpair_model3(const Vector<WordIndex>&es, const Vector<WordIndex>&fs, tmodel<COUNT, PROB>&tTable, amodel<PROB>&aTable, amodel<PROB>&dTable, nmodel<PROB>&nTable, double _p1, double _p0, void*x=0);
  WordIndex get_l()const
    {return l;}
  WordIndex get_m()const
    {return m;}
  const PROB&get_d(WordIndex i, WordIndex j)const
    {return d(i, j);}
  const PROB&get_a(WordIndex i, WordIndex j)const
    {return a(i, j);}
  const PROB&get_t(WordIndex i, WordIndex j)const
    {return t(i, j);}
  const PROB&get_fertility(WordIndex i, WordIndex f)const
    {return (f>=MAX_FERTILITY)?n(i, MAX_FERTILITY):n(i, f);}
  /*  bool exists(WordIndex i, WordIndex j)const
      {return e(i, j);}*/
  int modelnr()const{return 3;}
  LogProb scoreOfMove(const alignment&a, WordIndex new_i, WordIndex j, bool withDistortions=1)const;
  LogProb scoreOfSwap(const alignment&a, WordIndex j1, WordIndex j2, bool withDistortions=1)const ;
  LogProb _scoreOfMove(const alignment&a, WordIndex new_i, WordIndex j)const
    {
      alignment b(a);
      b.set(j, new_i);
      LogProb a_prob=prob_of_target_and_alignment_given_source(a);
      LogProb b_prob=prob_of_target_and_alignment_given_source(b);
      if( a_prob )
	return b_prob/a_prob;
      else if( b_prob )
	return 1e20;
      else
	return 1.0;
    }
  LogProb _scoreOfSwap(const alignment&a, WordIndex j1, WordIndex j2)const
    {
      alignment b(a);
      b.set(j1, a(j2));
      b.set(j2, a(j1));
      LogProb a_prob=prob_of_target_and_alignment_given_source(a);
      LogProb b_prob=prob_of_target_and_alignment_given_source(b);
      if( a_prob )
	return b_prob/a_prob;
      else if( b_prob )
	return 1e20;
      else
	return 1.0;
    }
  friend ostream&operator<<(ostream&out, const transpair_model3&m);
  LogProb prob_of_target_and_alignment_given_source(const alignment&al)const
    {
      LogProb total = 1.0 ;
      static const LogProb zero = 1E-299 ; 
      total *= pow(double(1-p1), m-2 * al.fert(0)) * pow(double(p1), al.fert(0));
      for (WordIndex i = 1 ; i <= al.fert(0) ; i++)
	total *= double(m - al.fert(0) - i + 1) / i ;
      for (WordIndex i = 1 ; i <= l ; i++)
	total *= get_fertility(i, al.fert(i)) * (LogProb) factorial(al.fert(i));
      for (WordIndex j = 1 ; j <= m ; j++)
	{
	  total*= get_t(al(j), j) ;
	  if (al(j))
	    total *= get_d(al(j), j);
	}
      return total?total:zero;
    }
};
#endif
