/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
/*--
transpair_model3: representation of a translation pair for model3 training
allowing for fast access (esp. to t table).

Franz Josef Och (30/07/99)
--*/
#include "transpair_model4.h"

LogProb transpair_model4::_scoreOfMove(const alignment&a, WordIndex new_i, WordIndex j)const
{
  alignment b(a);
  b.set(j, new_i);
  LogProb a_prob=prob_of_target_and_alignment_given_source(a);
  LogProb b_prob=prob_of_target_and_alignment_given_source(b);
  if( a_prob )
    return b_prob/a_prob;
  else if( b_prob )
    return 1e20;
  else
    return 1.0;
}
LogProb transpair_model4::_scoreOfSwap(const alignment&a, WordIndex j1, WordIndex j2)const 
{
  alignment b(a);
  b.set(j1, a(j2));
  b.set(j2, a(j1));
  LogProb a_prob=prob_of_target_and_alignment_given_source(a);
  LogProb b_prob=prob_of_target_and_alignment_given_source(b);
  if( a_prob )
    return b_prob/a_prob;
  else if( b_prob )
    return 1e20;
  else
    return 1.0;
}
//increasing efficiency: no copy of alignment (calc. everything incrementally)
LogProb transpair_model4::scoreOfMove(const alignment&a, WordIndex new_i, WordIndex j)const
{
  alignment b(a);
  b.set(j,new_i);
  LogProb change=transpair_model3::scoreOfMove(a,new_i,j,0);
  LogProb a_prob=prob_of_target_and_alignment_given_source(a,2);
  LogProb b_prob=prob_of_target_and_alignment_given_source(b,2);
  change*=b_prob/a_prob;
#ifdef VDEBUG
  LogProb realCosts=_scoreOfMove(a, new_i, j);
  if( (1.0-realCosts/change)>3e-2 )
    {
      cerr << "4Wrong costs of move: " << realCosts << " " << change << " " << (1.0-realCosts/change)<< endl;
      cerr << a << endl;
    }
#endif
  return change;
}
//increasing efficiency: no copy of alignment (calc. everything incrementally)
LogProb transpair_model4::scoreOfSwap(const alignment&a, WordIndex j1, WordIndex j2)const 
{
  alignment b(a);
  b.set(j1,a(j2));
  b.set(j2,a(j1));
  LogProb change=transpair_model3::scoreOfSwap(a,j1,j2,0);
  LogProb a_prob=prob_of_target_and_alignment_given_source(a,2);
  LogProb b_prob=prob_of_target_and_alignment_given_source(b,2);
  change*=b_prob/a_prob;
#ifdef VDEBUG
  PROB realCosts=_scoreOfSwap(a, j1, j2);
  if( (1.0-realCosts/change)>3e-2 )
    {
      cerr << "4Wrong costs of swap: " << realCosts << " " << change << " " << (1.0-realCosts/change) << endl;
      cerr << a <<endl;
    }
#endif
  return change;
}
