/*

EGYPT Toolkit for Statistical Machine Translation
Written by Yaser Al-Onaizan, Jan Curin, Michael Jahr, Kevin Knight, John Lafferty, Dan Melamed, David Purdy, Franz Och, Noah Smith, and David Yarowsky.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
USA.

*/
/*--
transpair_model4: representation of a translation pair for model3 training
allowing for fast access (esp. to t table).

Franz Josef Och (30/07/99)
--*/
#ifndef transpair_model4_h_fjo_defined
#define transpair_model4_h_fjo_defined
#include "Array2.h"
#include "defs.h"
#include "Vector.h"
#include "NTables.h"
#include "ATables.h"
#include "TTables.h"
#include "alignment.h"
#include "D4Tables.h"
#include "transpair_model3.h"

extern double factorial(int n);

class transpair_model4 : public transpair_model3
{
 private:
  const d4model&d4m;
 public:
  transpair_model4(const Vector<WordIndex>&es, const Vector<WordIndex>&fs, tmodel<COUNT, PROB>&tTable, amodel<PROB>&aTable, amodel<PROB>&dTable, nmodel<PROB>&nTable, double _p1, double _p0, const d4model*_d4m) 
    : transpair_model3(es, fs, tTable, aTable, dTable, nTable, _p1, _p0), d4m(*_d4m) {}
  LogProb scoreOfMove(const alignment&a, WordIndex new_i, WordIndex j)const;
  LogProb scoreOfSwap(const alignment&a, WordIndex j1, WordIndex j2)const ;
  LogProb _scoreOfMove(const alignment&a, WordIndex new_i, WordIndex j)const;
  LogProb _scoreOfSwap(const alignment&a, WordIndex j1, WordIndex j2)const ;
  int modelnr()const{return 4;}
  LogProb prob_of_target_and_alignment_given_source(const alignment&al, short distortionType=3)const
    {
      LogProb total = 1.0 ;
      static const LogProb almostZero = 1E-299 ; 
      if( distortionType&1 )
	{
	  total *= pow(double(1-p1), m-2 * al.fert(0)) * pow(double(p1), al.fert(0));
	  for (WordIndex i = 1 ; i <= al.fert(0) ; i++)
	    total *= double(m - al.fert(0) - i + 1) / i ;
	  for (WordIndex i = 1 ; i <= l ; i++)
	    total *= get_fertility(i, al.fert(i)) * (LogProb) factorial(al.fert(i));
	  for (WordIndex j = 1 ; j <= m ; j++)
	    total*= get_t(al(j), j) ;
	}
      if( distortionType&2 )
	{
	  for(WordIndex j=1;j<=m;j++)
	    if( al(j) )
	      if( al.get_head(al(j))==j)
		total*= d4m.getProb_first(j,al.get_center(al.prev_cept(al(j))),0,0);
	      else
		total*= d4m.getProb_second(j,al.prev_in_cept(j),0,0);
	}
      return total?total:almostZero;
    }
};
#endif
